%{
#ifndef lint
static char *rcsid = "$Header: /usr/src/local/awl/RCS/awl.y,v 2.6 90/04/19 20:04:34 jkh Exp $";
#endif

/*
 *
 *                     Copyright 1989
 *                    Jordan K. Hubbard
 *
 *		  PCS Computer Systeme, GmbH.
 *	             Munich, West Germany
 *
 *
 * This file is part of AWL.
 *
 * AWL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * AWL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AWL; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

/*
 * This is awl's main yacc parser.
 *
 * $Log:	awl.y,v $
 * Revision 2.6  90/04/19  20:04:34  jkh
 * Alpha checkin.
 * 
 * Revision 2.5  90/03/30  22:01:33  jkh
 * Drastically re-arranged the opcodes.
 * 
 * Revision 2.4  90/03/30  20:13:23  jkh
 * Got switch/case working. Fixed list constants.
 * 
 * Revision 2.3  90/03/29  18:25:26  jkh
 * Added first cut at switch/case.
 * 
 * Revision 2.2  90/03/27  09:13:21  jkh
 * Added PIXMAP builtin type.
 * 
 * Revision 2.1  90/03/26  13:11:45  jkh
 * Removed rcsclean from allclean rule; overkill.
 * 
 * Revision 2.0  90/03/26  01:42:52  jkh
 * pre-beta check-in
 *
 */

#include "AwlP.h"

/* For break/continue statements */
Local Symbol BreakList = NULL, ContinueList = NULL;

/* For switch statements */
#define NO_SWITCH	0
#define IN_SWITCH	1
#define SEEN_DEFAULT	2
Local Symbol CaseHeaderList = NULL, CaseBodyList = NULL;
Local int CaseState = NO_SWITCH;

/* For local variables */
Local Symbol CurrentFunc = NULL;

/* For list constants */
String *ListKludge;
int ListKludgeSize, ListKludgeMax;

/* Misc */
Local Datum d1;

Export int yyerror();
Export void yywarn();

Import AwlWidget _currW;

#define YYKOTZ(msg, arg) { if (yyerror((msg), (arg))) return(1); }

%}

%union
{
     char		bval;
     WidgetClass	cval;
     int		ival;
     FTYPE		fval;
     String		sval;
     String	       *lval;
     Symbol		symb;
     Datum		data;
}

/* From K&R, (plus the special operators) */
%left	COMMA
%right	EQ ME PE TE DE RE XE AE OE CE SLE SRE
%right	QST COL
%left	LOR
%left	LAND
%left	OR
%left	XOR
%left	AND COPY NUKE
%left	EE NE
%left	LT LE GT GE
%left	SL SR
%left	PLUS MINUS
%left	MULT DIV MOD
%right	NOT ADDR COMP PP MM UMINUS CAST RANGE
%left	MBR
%left	DOT LBRK RBRK

/*
 * The general instruction set (minus most ALU ops, which were def'd above)
 * The order of these puppies is important! Don't mess with it without
 * reading the helpful comments in eval.c where opcode_table is declared.
 */
%token	ALU CALL JUMP JUMPZ POP RTS ARES CLOS DUP HALT LINK OPEN PUSH

/* ALU ops not used directly by expressions. _LASTALU must always be last */
%token	ASOC SUB TRI _LASTALU

/* Registers and internal types */
%token	_PC _SP _FP
%token	FUNC_UNDEF FUNC_ACTION FUNC_BUILTIN FUNC_CALLBACK FUNC_LAYOUT
%token	FUNC_USER FUNC_EXTERN FILEDESC FILEPDESC RESOURCE SYMBOL RANGE

/* public types */
%token	ANY CADDR CALLBACK CHAR FADDR FLOAT IADDR INT LAYOUT LIST PIXMAP SADDR
%token	STRING TRANSLATION WCLASS WIDGET

/* classes */
%token ALLOCATED CONSTANT DATA FUNCTION INDIRECT LABEL STATIC

/* keywords and separators */
%token	';' '{' '}' '(' ')'
%token	IF ELSE WHILE DO FOR UNTIL BREAK CONTINUE RETURN GOTO SWITCH
%token	CASE DEFAULT

%token	GLOBAL LOCAL

%nonassoc IF
%nonassoc ELSE

%token	<ival>	INT
%token	<fval>	FLOAT
%token	<bval>	CHAR
%token	<sval>	STRING
%token	<sval>	IDENTIFIER
%token  <cval>	WIDGET_TYPE

%type	<data>	char_data int_data float_data string_data
%type	<data>	range_data list_data wclass_data

%type	<data>	widget_data
%type	<ival>	numeric do_prefix
%type	<sval>	string ident
%type	<symb>	newparm parm parameter_list opt_parm_list
%type	<symb>	function_start
%type	<symb>	if_prefix else_clause
%type	<symb>	while_prefix while_loop until_loop
%type	<ival>	f_cond
%type	<symb>	f_lb1 f_lb2 f_last
%type	<lval>	list

%start	declaration

%%

global_declaration
     :	GLOBAL global_body semi
     ;

global_body
     :	global_keyword
     |	global_body COMMA global_keyword
	{ yyerrok; }
     |	error
	{ YYKOTZ("Missing or invalid symbol name for global", 0); }
     |	global_body error
	{ YYKOTZ("Unexpected symbol following global definition", 0); }
     |	global_body error global_keyword
	{ YYKOTZ("Missing or invalid separator", 0); }
     |	global_body COMMA error
	{ YYKOTZ("Missing or invalid symbol list for global", 0); }
     ;

global_keyword
     :	ident
	{
	     Symbol tmp;

	     tmp = symbol_new(_currW, $1, STRING);
	     symbol_aobj(tmp) = new_aobj($1);
	     symbol_add(tmp);
	}
     |	ident EQ CHAR
	{
	     Symbol tmp = symbol_new(_currW, $1, CHAR);
	     symbol_char(tmp) = $3;
	     symbol_add(tmp);
	     XtFree($1);
	}
     |	ident EQ INT
	{
	     Symbol tmp = symbol_new(_currW, $1, INT);
	     symbol_int(tmp) = $3;
	     symbol_add(tmp);
	     XtFree($1);
	}
     |	ident EQ FLOAT
	{
	     Symbol tmp = symbol_new(_currW, $1, FLOAT);
	     symbol_float(tmp) = $3;
	     symbol_add(tmp);
	     XtFree($1);
	}
     |	ident EQ range_data
	{
	     Symbol tmp = symbol_new(_currW, $1, RANGE);
	     symbol_data(tmp) = $3;
	     symbol_add(tmp);
	     XtFree($1);
	}
     |	ident EQ string
	{
	     Symbol tmp = symbol_new(_currW, $1, STRING);
	     symbol_aobj(tmp) = new_aobj($3);
	     symbol_add(tmp);
	     XtFree($1);
	}
     |	ident EQ list
	{
	     Symbol tmp = symbol_new(_currW, $1, LIST);
	     symbol_aobj(tmp) = new_aobj($3);
	     symbol_add(tmp);
	     XtFree($1);
	}
     |	ident lpar rpar
	{
	     symbol_add_function(_currW, $1);
	     XtFree($1);
	}
     |	ident LBRK INT RBRK
	{
	     Symbol tmp;

	     tmp = symbol_new(_currW, $1, STRING);
	     symbol_class(tmp) = ALLOCATED;
	     symbol_aobj(tmp) = new_aobj(XtMalloc($3));
	     bzero(aobj_ptr(symbol_aobj(tmp)), $3);
	     symbol_add(tmp);
	     XtFree($1);
	}
     |	ident EQ error
	{ YYKOTZ("Missing or invalid initializer for global", 0); }
     |	ident LBRK error RBRK
	{ YYKOTZ("Invalid or missing size for global array type", 0); }
     ;

declaration
     :	function_declaration
     |	declaration function_declaration
	{ yyerrok; }
     |	global_declaration
     |	declaration global_declaration
	{ yyerrok; }
     |	error
	{ YYKOTZ("Missing or invalid function/global declaration", 0); }
     |	declaration error
	{ YYKOTZ("Invalid function/global declaration", 0); }
     ;

function_declaration
     :	function_start compound_statement
	{
	     Function fp = symbol_function($1);

	     /*
	      * emit an RTS in case there was no return. If there was,
	      * it will be automatically ignored by do_emit().
	      */
	     do_emit(_currW, RTS, FalseDatum);
	     function_end(fp) = awl_lastpc(_currW);
	     if (awl_showasm(_currW)) {
		  fprintf(awl_err(_currW),
			  "\nCode for %s:\n", symbol_name($1));
		  disasm(_currW, function_uaddr(fp), function_end(fp));
	     }
	     symbol_set_locals(NULL);
	     CurrentFunc = NULL;
	}
     ;

function_start
     :	ident opt_parm_list
	{
	     int offset;
	     Symbol tmp;
	     Address addr;

	     tmp = $2;
	     offset = 1;
	     while(tmp) {
		  symbol_int(tmp) = -(offset++);
		  tmp = symbol_next(tmp);
	     }
	     symbol_set_locals($2);
	     addr_pc(addr) = awl_lastpc(_currW);
	     $$ = symbol_def_function(_currW, $1, FUNC_USER,
				      offset - 1, addr, TRUE);
	     CurrentFunc = $$;
	     XtFree($1);
	}
     ;

opt_parm_list
     :	lpar rpar
	{
	     $$ = NULL;
	     yyerrok;
	}
     |	lpar parameter_list rpar
	{
	     $$ = $2;
	     yyerrok;
	}
     |	error
	{ YYKOTZ("Missing or invalid parameter list", 0); }
     ;

newparm
     :	ident
	{
	     $$ = symbol_new(_currW, $1, _FP);
	     XtFree($1);
	}
     ;

parm
     :	newparm
     |	newparm
	{
	     datum_symbol(d1) = $1;
	     do_push(_currW, SYMBOL, d1);
	}
	EQ expr
	{ do_emit(_currW, POP, PopDupNone); }
     ;

parameter_list
     :	parm
     |	parm COMMA parameter_list
	{
	     symbol_next($1) = $3;
	     yyerrok;
	}
     |	error
	{ YYKOTZ("Missing or invalid parameter in parameter list", 0); }
     ;

widget_data
     :	widget_decl
	{ $$ = FalseDatum; }
     ;

compound_statement
     :	lbra statements rbra
     ;

statements
     :	statement
     |	statements statement
	{ yyerrok; }
     |	error
	{ YYKOTZ("Missing or invalid statement", 0); }
     |	statements error
	{ YYKOTZ("Invalid statement", 0); }
     ;

statement
     :	semi	/* null statement */
     |	discarding_expr
     |	compound_statement
     |	BREAK semi
	{
	     if (BreakList) {
		  datum_symbol(d1) = BreakList;
		  do_emit(_currW, JUMP, d1);
	     }
	     else
		  YYKOTZ("Break not inside loop or switch.", 0)
	}
     |	CONTINUE semi
	{
	     if (ContinueList) {
		  datum_symbol(d1) = ContinueList;
		  do_emit(_currW, JUMP, d1);
	     }
	     else
		  YYKOTZ("Continue not inside loop.", 0)
	}
     |	RETURN semi
	{ do_emit(_currW, RTS, FalseDatum); }
     |	RETURN expression semi
	{ do_emit(_currW, RTS, TrueDatum); }
     |	ident COL
	{
	     Symbol tmp;

	     if ((tmp = symbol_find(_currW, $1)) != NULL) {
		  if (symbol_class(tmp) != LABEL) /* illegal */
		       YYKOTZ("Attempt to redefine %s as label.", $1)
		  else if (symbol_int(tmp) != UNDEF_ADDR)
		       YYKOTZ("Attempted redefinition of label %s.", $1)
	     }
	     else {
		  tmp = symbol_new(_currW, $1, INT);
		  symbol_class(tmp) = LABEL;
	     }
	     symbol_int(tmp) = awl_lastpc(_currW);
	     symbol_add(tmp);
	     debug(_currW, "label '%s' declared", $1, awl_lastpc(_currW));
	     XtFree($1);
	}
     |	GOTO ident semi
	{
	     Symbol tmp;

	     if ((tmp = symbol_find(_currW, $2)) != NULL) {
		  if (symbol_class(tmp) != LABEL)
		       YYKOTZ("Illegal target for GOTO: %s", $2)
		  else {
		       datum_symbol(d1) = tmp;
		       do_emit(_currW, JUMP, d1);
		  }
	     }
	     else {
		  tmp = symbol_new(_currW, $2, INT);
		  symbol_class(tmp) = LABEL;
		  symbol_type(tmp) = INT;
		  symbol_int(tmp) = UNDEF_ADDR;
		  symbol_add(tmp);
		  datum_symbol(d1) = tmp;
		  do_emit(_currW, JUMP, d1);
		  debug(_currW, "GOTO to forward label %s", $2);
	     }
	     XtFree($2);
	}
     |	LOCAL parameter_list semi
	{
	     Symbol tmp = $2;

	     if (!CurrentFunc)
		  YYKOTZ("LOCAL keyword not in any function", 0)
	     else {
		  while (tmp) {
		       symbol_int(tmp) =
			    ++function_nlocals(symbol_function(CurrentFunc));
		       tmp = symbol_next(tmp);
		  }
		  symbol_set_locals($2);
	     }
	}
     |	global_declaration
     |	if_prefix statement
	{ symbol_int($1) = awl_lastpc(_currW); }
     |	if_prefix statement else_clause
	{ symbol_int($1) = awl_lastpc(_currW); }
     	statement
	{ symbol_int($3) = awl_lastpc(_currW); }
     |	while_prefix while_loop statement
	{
	     datum_symbol(d1) = $1;
	     do_emit(_currW, JUMP, d1);
	     symbol_int($2) = awl_lastpc(_currW);
	     ContinueList = symbol_next(ContinueList);
	     BreakList = symbol_next(BreakList);
	     yyerrok;
	}
     |	while_prefix error
	{ YYKOTZ("Missing or invalid expression for while", 0); }
     |	do_prefix statement until_loop
	{
	     symbol_int($3) = $1;
	     ContinueList = symbol_next(ContinueList);
	     BreakList = symbol_next(BreakList);
	     yyerrok;
	}
     |	do_prefix error
	{ YYKOTZ("Missing or invalid statement for do", 0); }
     |	FOR lpar f_init f_lb0 f_cond f_lb1 f_itr f_lb2 f_last statement
	{
	     /* make the jump to f_lb1 */
	     datum_symbol(d1) = $6;
	     do_emit(_currW, JUMP, d1);
	     /* backpatch the f_cond JUMP */
	     datum_symbol(awl_text(_currW)[$5]) = $8;
	     symbol_int($9) = awl_lastpc(_currW);
	     ContinueList = symbol_next(ContinueList);
	     BreakList = symbol_next(BreakList);
	     yyerrok;
	}
     |	FOR error
	{ YYKOTZ("Invalid syntax in for loop", 0); }
     |	switch_prefix statement
	{
	     if (CaseHeaderList) {
		  symbol_int(CaseHeaderList) = awl_lastpc(_currW);
		  CaseHeaderList = symbol_next(CaseHeaderList);
	     }
	     if (CaseBodyList) {
		  symbol_int(CaseBodyList) = awl_lastpc(_currW);
		  CaseBodyList = symbol_next(CaseBodyList);
	     }
	     symbol_int(BreakList) = awl_lastpc(_currW);
	     BreakList = symbol_next(BreakList);
	     CaseState = NO_SWITCH;
	     /* Get rid of the last dup'd expression */
	     do_discard(_currW);
	     yyerrok;
	}
     |	switch_prefix error
	{ YYKOTZ("Missing body for switch", 0); }
     |	case_prefix expression COL
	{
	     Symbol tmp;

	     /* Check to see if case is true. If not, jump to next case */
	     do_alu(_currW, EE);
	     datum_symbol(d1) = CaseHeaderList;
	     do_emit(_currW, JUMPZ, d1);
	     if (CaseBodyList) {
		  symbol_int(CaseBodyList) = awl_lastpc(_currW);
		  CaseBodyList = symbol_next(CaseBodyList);
	     }
	     tmp = do_new_label(_currW, UNDEF_ADDR);
	     symbol_next(tmp) = CaseBodyList;
	     CaseBodyList = tmp;
	     yyerrok;
	}
     |	case_prefix error
	{ YYKOTZ("Invalid syntax for case label", 0); }
     |	case_prefix error COL
	{ YYKOTZ("Invalid expression for case label", 0); }
     |	DEFAULT COL
	{
	     Symbol tmp;

	     if (CaseState == NO_SWITCH)
		  YYKOTZ("default: case not in switch", 0)
	     else if (CaseState == SEEN_DEFAULT)
		  YYKOTZ("More than one 'default:' clause in switch", 0)
	     else {
		  CaseState = SEEN_DEFAULT;
		  if (CaseHeaderList) {
		       symbol_int(CaseHeaderList) = awl_lastpc(_currW);
		       CaseHeaderList = symbol_next(CaseHeaderList);
		  }
		  if (CaseBodyList) {
		       symbol_int(CaseBodyList) = awl_lastpc(_currW);
		       CaseBodyList = symbol_next(CaseBodyList);
		  }
		  tmp = do_new_label(_currW, UNDEF_ADDR);
		  symbol_next(tmp) = CaseHeaderList;
		  CaseHeaderList = tmp;
		  tmp = do_new_label(_currW, awl_lastpc(_currW));
		  symbol_next(tmp) = CaseBodyList;
		  CaseBodyList = tmp;
	     }
	}
     ;

if_prefix
     :	IF lpar expression rpar
	{
	     $$ = do_new_label(_currW, UNDEF_ADDR);
	     datum_symbol(d1) = $$;
	     do_emit(_currW, JUMPZ, d1);
	     yyerrok;
	}
     |	IF error
	{ YYKOTZ("Missing or invalid if", 0); }
     ;

else_clause
     :	ELSE
	{
	     $$ = do_new_label(_currW, UNDEF_ADDR);
	     datum_symbol(d1) = $$;
	     do_emit(_currW, JUMP, d1);
	}
     ;

while_prefix
    :	WHILE
	{
	     $$ = do_new_label(_currW, awl_lastpc(_currW));
	     symbol_next($$) = ContinueList;
	     ContinueList = $$;
	}
     ;

while_loop
     :	lpar expression rpar
	{
	     $$ = do_new_label(_currW, UNDEF_ADDR);
	     datum_symbol(d1) = $$;
	     do_emit(_currW, JUMPZ, d1);
	     symbol_next($$) = BreakList;
	     BreakList = $$;
	}
     ;

do_prefix
     :	DO
	{
	     Symbol tmp;

	     $$ = awl_lastpc(_currW);
	     tmp = do_new_label(_currW, $$);
	     symbol_next(tmp) = ContinueList;
	     ContinueList = tmp;
	     tmp = do_new_label(_currW, UNDEF_ADDR);
	     symbol_next(tmp) = BreakList;
	     BreakList = tmp;
	}
     ;

until_loop
     :	UNTIL lpar expression rpar
	{
	     $$ = datum_symbol(d1) = do_new_label(_currW, UNDEF_ADDR);
	     do_emit(_currW, JUMPZ, d1);
	     symbol_int(BreakList) = awl_lastpc(_currW);
	     yyerrok;
	}
     |	UNTIL error
	{ YYKOTZ("Missing or invalid expression for until", 0); }
     ;

f_init
     :	optional_expression
     ;

f_lb0
     :	semi
	{
	     Symbol tmp;

	     /* create label pointing at f_cond */
	     tmp = do_new_label(_currW, awl_lastpc(_currW));
	     symbol_next(tmp) = ContinueList;
	     ContinueList = tmp;
	}
     ;

f_cond
     :	/* nichts */
	{
	     do_push(_currW, INT, TrueDatum);
	     /* point us at the place to patch the jump later */
	     $$ = awl_lastpc(_currW) + 1;
	     datum_symbol(d1) = NULL;
	     do_emit(_currW, JUMP, d1);
	}
     |	expression
	{
	     /* point us at the place to patch the jump later */
	     $$ = awl_lastpc(_currW) + 1;
	     datum_symbol(d1) = NULL;
	     do_emit(_currW, JUMP, d1);
	}
     ;

f_lb1
     :	semi
	{ $$ = do_new_label(_currW, awl_lastpc(_currW)); }
     ;

f_itr
     :	optional_expression
	{
	     /*
	      * we know that the ContinueList will be pointing at lb0
	      * (f_cond) so we just use that rather than backpatching.
	      */
	     datum_symbol(d1) = ContinueList;
	     do_emit(_currW, JUMP, d1);
	}
     ;

f_lb2
     :	/* nada */
	{ $$ = do_new_label(_currW, awl_lastpc(_currW)); }
     ;

f_last
     :	rpar
	{
	     /* create dummy label for jump to end of for */
	     $$ = do_new_label(_currW, UNDEF_ADDR);
	     datum_symbol(d1) = $$;
	     do_emit(_currW, JUMPZ, d1);
	     symbol_next($$) = BreakList;
	     BreakList = $$;
	}
     ;

switch_prefix
     :	SWITCH lpar expression rpar
	{
	     Symbol tmp;

	     tmp = do_new_label(_currW, UNDEF_ADDR);
	     symbol_next(tmp) = BreakList;
	     BreakList = tmp;
	     CaseState = IN_SWITCH;
	     yyerrok;
	}
     |	SWITCH error
	{ YYKOTZ("Missing or invalid expression for switch", 0); }
     ;

case_prefix
     :	CASE
	{
	     Symbol tmp;

	     if (CaseState == NO_SWITCH)
		  YYKOTZ("Case not in switch", 0)
	     else if (CaseState == SEEN_DEFAULT)
		  yywarn("Case after 'default:' clause will never be reached");
	     else {
		  if (CaseBodyList)
		       do_optional_jump(_currW, CaseBodyList);
		  if (CaseHeaderList) {
		       symbol_int(CaseHeaderList) = awl_lastpc(_currW);
		       CaseHeaderList = symbol_next(CaseHeaderList);
		  }
		  tmp = do_new_label(_currW, UNDEF_ADDR);
		  symbol_next(tmp) = CaseHeaderList;
		  CaseHeaderList = tmp;
		  do_emit(_currW, DUP);
	     }
	}
     ;

expression
     :	expr
     |	expression COMMA expr
	{ yyerrok; }
     ;

optional_expression
     :	/* rien */
     |	expression
	{ do_discard(_currW); }
     ;

discarding_expr
     :	expression semi
	{ do_discard(_currW); }
     ;

l1_expr
     :	identifier
     |	funcall
     |	l1_expr DOT ident
	{
	     do_push(_currW, STRING, new_aobj($3));
	     do_alu(_currW, DOT);
	}
     |	l1_expr LBRK expr RBRK
	{ do_alu(_currW, ASOC); }
     |	l1_expr DOT error
	{ YYKOTZ("Missing or invalid member reference", 0); }
     |	l1_expr LBRK error RBRK
	{ YYKOTZ("Missing or invalid array reference", 0); }
     |	l1_expr error
	{ YYKOTZ("Missing or invalid identifier", 0); }
     ;

dupl1expr
     :	l1_expr
	{ do_emit(_currW, DUP); }
     ;

l2_expr
     :	l1_expr
     |	l1_expr EQ expr
	{ do_emit(_currW, POP, PopDupValue); }
     |	PP dupl1expr
	{
	     do_push(_currW, INT, TrueDatum); /* true = convenient 1 */
	     do_alu(_currW, PLUS);
	     do_emit(_currW, POP, PopDupValue);
	}
     |	MM dupl1expr
	{
	     do_push(_currW, INT, TrueDatum);
	     do_alu(_currW, SUB);
	     do_emit(_currW, POP, PopDupValue);
	}
     |	dupl1expr PP
	{
	     do_push(_currW, INT, TrueDatum);
	     do_alu(_currW, PLUS);
	     do_emit(_currW, POP, PopDupTarget);
	}
     |	dupl1expr MM
	{
	     do_push(_currW, INT, TrueDatum);
	     do_alu(_currW, SUB);
	     do_emit(_currW, POP, PopDupTarget);
	}
     |  dupl1expr PE expr
	{
	     do_alu(_currW, PLUS);
	     do_emit(_currW, POP, PopDupValue);
	}
     |  dupl1expr ME expr
	{
	     do_alu(_currW, SUB);
	     do_emit(_currW, POP, PopDupValue);
	}
     |  dupl1expr TE expr
	{
	     do_alu(_currW, MULT);
	     do_emit(_currW, POP, PopDupValue);
	}
     |  dupl1expr DE expr
	{
	     do_alu(_currW, DIV);
	     do_emit(_currW, POP, PopDupValue);
	}
     |  dupl1expr RE expr
	{
	     do_alu(_currW, MOD);
	     do_emit(_currW, POP, PopDupValue);
	}
     |  dupl1expr SLE expr
	{
	     do_alu(_currW, SL);
	     do_emit(_currW, POP, PopDupValue);
	}
     |  dupl1expr SRE expr
	{
	     do_alu(_currW, SR);
	     do_emit(_currW, POP, PopDupValue);
	}
     |  dupl1expr XE expr
	{
	     do_alu(_currW, XOR);
	     do_emit(_currW, POP, PopDupValue);
	}
     |  dupl1expr OE expr
	{
	     do_alu(_currW, OR);
	     do_emit(_currW, POP, PopDupValue);
	}
     |  dupl1expr AE expr
	{
	     do_alu(_currW, AND);
	     do_emit(_currW, POP, PopDupValue);
	}
     |  dupl1expr CE expr
	{
	     do_alu(_currW, COMP);
	     do_emit(_currW, POP, PopDupValue);
	}
     ;

expr
     :	l2_expr
     |	lpar expr rpar
     |	char_data			{ do_push(_currW, CHAR, $1);	}
     |	int_data			{ do_push(_currW, INT, $1);	}
     |  float_data			{ do_push(_currW, FLOAT, $1);	}
     |  string_data			{ do_push(_currW, STRING, $1);	}
     |	list_data			{ do_push(_currW, LIST, $1);	}
     |	range_data			{ do_push(_currW, RANGE, $1);	}
     |	wclass_data			{ do_push(_currW, WCLASS, $1);	}
     |	widget_data			{ do_push(_currW, LAYOUT, $1);	}
     |	expr MBR expr			{ do_alu(_currW, MBR);		}
     |	expr PLUS expr			{ do_alu(_currW, PLUS);		}
     |	expr MINUS expr			{ do_alu(_currW, SUB);		}
     |	expr MULT expr			{ do_alu(_currW, MULT);		}
     |	expr DIV expr			{ do_alu(_currW, DIV);		}
     |  expr MOD expr			{ do_alu(_currW, MOD);		}
     |  expr LT expr			{ do_alu(_currW, LT);		}
     |	expr GT expr			{ do_alu(_currW, GT);		}
     |	expr LE expr			{ do_alu(_currW, LE);		}
     |	expr GE expr			{ do_alu(_currW, GE);		}
     |	expr EE expr			{ do_alu(_currW, EE);		}
     |	expr NE expr			{ do_alu(_currW, NE);		}
     |	expr SL expr			{ do_alu(_currW, SL);		}
     |	expr SR expr			{ do_alu(_currW, SR);		}
     |	expr OR expr			{ do_alu(_currW, OR);		}
     |  expr AND expr			{ do_alu(_currW, AND);		}
     |	expr XOR expr			{ do_alu(_currW, XOR);		}
     |	expr LAND expr			{ do_alu(_currW, LAND);		}
     |	expr LOR expr			{ do_alu(_currW, LOR);		}
     |	expr QST expr COL expr		{ do_alu(_currW, TRI);		}
     |	lpar expr rpar expr %prec CAST	{ do_alu(_currW, CAST);		}
     |	MINUS expr %prec UMINUS		{ do_alu(_currW, MINUS);	}
     |	AND expr %prec ADDR		{ do_alu(_currW, ADDR);		}
     |	COPY expr			{ do_alu(_currW, COPY);		}
     |	NUKE expr			{ do_alu(_currW, NUKE);		}
     |	COMP expr			{ do_alu(_currW, COMP);		}
     |	NOT expr 			{ do_alu(_currW, NOT);		}
     ;

identifier
     :	ident
	{
	     Symbol tmp;

	     /*
	      * Unknown identifiers are created automatically as string syms.
	      */
	     if ((tmp = symbol_find(_currW, $1)) == NULL) {
		  tmp = symbol_new(_currW, $1, STRING);
		  symbol_aobj(tmp) = new_aobj($1);
		  symbol_add(tmp);
	     }
	     datum_symbol(d1) = tmp;
	     do_push(_currW, SYMBOL, d1);
	}
     ;

funcall
     :	ident params
	{
	     Symbol tmp;

	     if ((tmp = symbol_find(_currW, $1)) == NULL) 
		  tmp = symbol_add_function(_currW, $1); /* forward ref */
	     datum_symbol(d1) = tmp;
	     do_emit(_currW, CALL, d1);
	     XtFree($1);
	}
     ;

params
     :	s_parm e_parm
     |	s_parm expression e_parm
     ;

s_parm
     :	lpar
	{ do_emit(_currW, LINK); }
     ;

e_parm
     :	rpar
     ;

widget_decl
     :	widget_open widget_body
	{ do_emit(_currW, CLOS); }
     ;

widget_body
     :	lbra widget_line rbra
     |	semi
     ;

widget_open
     :	widget_type widget_name
	{ do_emit(_currW, OPEN); }
     ;

widget_type
     :	wclass_data
	{ do_push(_currW, WCLASS, $1); }
     |	l1_expr
     ;

widget_name
     :	string_data
	{ do_push(_currW, STRING, $1); }
     |	l1_expr
     ;

widget_line
     :	resource
     |	widget_line resource
     |	widget_decl
     |	widget_line widget_decl
     |	work_proc
     |	widget_line work_proc
     ;

resource
     :	string_data expr semi
	{
	     do_push(_currW, STRING, $1);
	     do_emit(_currW, ARES);
	}
     ;

work_proc
     :	funcall semi
	{ do_discard(_currW); }
     ;

wclass_data
     :	WIDGET_TYPE
	{ datum_wclass($$) = $1; }
     ;

char_data
     :	CHAR
	{ datum_char($$) = $1; }
     ;

int_data
     :	INT
	{ datum_int($$) = $1; }
     ;

float_data
     :	FLOAT
	{ datum_float($$) = $1; }
     ;

string_data
     :	string
	{ datum_aobj($$) = new_aobj($1); }
     ;

list_data
     :	list
	{ datum_aobj($$) = new_aobj($1); }
     ;

range_data
     :	numeric RANGE numeric
	{
	     range_min(datum_range($$)) = (short)$1;
	     range_max(datum_range($$)) = (short)$3;
	}
     ;

numeric
     :	INT
     |	CHAR
	{ $$ = (int)$1; }
     |	FLOAT
	{ $$ = (int)$1; }
     ;

string
     :	STRING
     |	STRING MULT INT
	{
	     $$ = string_multiply($1, $3);
	     debug(_currW, "parse: string * freeing %s", $1);
	     XtFree($1);
	     yyerrok;
	}
     ;

ident
     :	IDENTIFIER
     ;

list
     :	list_start list_elements list_end
	{ $$ = ListKludge; }
     ;

list_start
     :	lbra
	{
	     ListKludgeMax = 5;
	     ListKludgeSize = 0;
	     ListKludge = (String *)XtMalloc(sizeof(String *) * 6);
	}
     ;

list_elements
     :	string
	{
	     if (ListKludgeSize == ListKludgeMax) {
		  ListKludgeMax += 5;
		  ListKludge =(String *)XtRealloc(ListKludge, sizeof(String *)
						  * (ListKludgeMax + 1));
	     }
	     ListKludge[ListKludgeSize++] = $1;
	}
     |	list_elements COMMA string
	{
	     if (ListKludgeSize == ListKludgeMax) {
		  ListKludgeMax += 5;
		  ListKludge =(String *)XtRealloc(ListKludge, sizeof(String *)
						  * (ListKludgeMax + 1));
	     }
	     ListKludge[ListKludgeSize++] = $3;
	}
     ;

list_end
     :	rbra
	{ ListKludge[ListKludgeSize] = NULL; }
     ;

lpar
     :	'('
     ;

rpar
     :	')'
	{ yyerrok; }
     ;

lbra
     :	'{'
     ;

rbra
     :	'}'
	{ yyerrok; }
     ;

semi
     :	';'
     ;

%%

/* Additional C support routines. */

/* The to-be-expected error routine */
Export int yyerror(s, n)
String s;
Generic n;
{
     Local int err_cnt = 0;
     char msg1[255], msg2[512];

     if (err_cnt++ > 5) {
	  sprintf(msg1, "%s:%d: Too many errors, aborting..\n",
		  yy_file, yy_line);
	  XtError(msg1);
	  err_cnt = 0;
	  return 1;
     }
     sprintf(msg1, s, n);
     if (yy_file)
	  sprintf(msg2, "%s:%d: %s\n", yy_file,
		  yy_line ? yy_line - 1 : 0, msg1);
     else
	  sprintf(msg2, "<stdin>:%d: %s\n",
		  yy_line ? yy_line - 1 : 0, msg1);
     XtError(msg2);
     return 0;
}

/* Like yyerror, but only a warning */
Export void yywarn(s, n)
String s;
Generic n;
{
     char msg1[255], msg2[512];

     if (!awl_nowarn(_currW)) {
	  sprintf(msg1, s, n);
	  if (yy_file)
	       sprintf(msg2, "%s:%d: %s\n", yy_file,
		       yy_line ? yy_line - 1 : 0, msg1);
	  else
	       sprintf(msg2, "<stdin>:%d: %s\n", 
		       yy_line ? yy_line - 1 : 0, msg1);
	  XtWarning(msg2);
     }
}
