#ifndef lint
static char *rcsid = "$Header: /usr/src/local/awl/RCS/dasm.c,v 2.1 90/03/30 22:00:03 jkh Exp $";
#endif

/*
 *
 *                     Copyright 1989
 *                    Jordan K. Hubbard
 *
 *		  PCS Computer Systeme, GmbH.
 *	             Munich, West Germany
 *
 *
 * This file is part of AWL.
 *
 * AWL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * AWL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AWL; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*
 * Functions for printing and disassembling portions of the text
 * segment. Also contains some general instruction info functions for
 * eval.c.
 *
 * $Log:	dasm.c,v $
 * Revision 2.1  90/03/30  22:00:03  jkh
 * Shuffled some stuff into evalutils.c
 * 
 * Revision 2.0  90/03/26  01:44:19  jkh
 * pre-beta check-in
 *
 */

#include "AwlP.h"
#include "y.tab.h"

/*
 * Return the print name for an opcode. Since the opcode table is always
 * sorted, we can bsearch it.
 */
Export String op_name(insn)
int insn;
{
     register int i, top = 0, bottom = opcode_table_size;
     char *name = NULL;

     while (top <= bottom && name == NULL) {
	  i = (top + bottom) / 2;
	  if (opcode_table[i].insn == insn)
	       name = opcode_table[i].name;
	  else if (opcode_table[i].insn < insn)
	       top = i + 1;
	  else
	       bottom = i - 1;
     }
     return(name);
}

/*
 * Return the print name for an ALU opcode. Since the table is declared
 * in order of increasing precendence (as is the grammer), we can bsearch
 * it.
 */
Export String alu_name(op)
int op;
{
     register int i, top = 0, bottom = alu_table_size;
     char *name = NULL;

     while (top <= bottom && name == NULL) {
	  i = (top + bottom) / 2;
	  if (alu_opcodes[i].insn == op)
	       name = alu_opcodes[i].name;
	  else if (alu_opcodes[i].insn < op)
	       top = i + 1;
	  else
	       bottom = i - 1;
     }
     if (name)
	  return(name);
     else
	  return("???");
}

/* Return the print name for a register */
Export String reg_name(op)
int op;
{
     if (op == _PC)
	  return("$PC");
     else if (op == _SP)
	  return("$SP");
     else if (op == _FP)
	  return("$FP");
     else
	  return("$??");
}

/* Return the print name for a type. */
Export Inline String type_name(typ)
short typ;
{
     String name;

     if (is_register(typ))
	  name = reg_name(typ);
     else
	  name = map_token_to_resword(typ, T_TYPE | T_INTERN_TYPE);
     if (!name)
	  name = "???";
     return(name);
}

Export Inline String class_name(class)
short class;
{
     String name;

     if ((name = map_token_to_resword(class, T_CLASS)) == NULL)
	  name = "???";
     return(name);
}
     
Local int show_labels(sym, data)
Symbol sym;
int *data;
{
     if (symbol_class(sym) == LABEL && symbol_int(sym) == *data)
	  fprintf(awl_err(symbol_root(sym)), "%s:\n", symbol_name(sym));
     return(TRUE);
}

/*
 * Find the line table entry for a given PC, if possible.
 */
Export LTabEnt *ltab_for_pc(aw, pc)
AwlWidget aw;
int pc;
{
     LTabEnt *ret = NULL;

     if (awl_ltab(aw)) {
	  register int i, top, bottom;
	  int line;

	  bottom = awl_ltabcurr(aw);
	  top = line = 0;

	  while (top <= bottom && ret == NULL) {
	       i = (top + bottom) / 2;
	       if (ltab_start(awl_ltab(aw)[i]) <= pc &&
		   ltab_last(awl_ltab(aw)[i]) >= pc)
		    ret = awl_ltab(aw) + i;
	       else if (ltab_last(awl_ltab(aw)[i]) < pc)
		    top = i + 1;
	       else
		    bottom = i - 1;
	  }
     }
     return(ret);
}

/* to make a couple of expressions more readable later */
#define SFDS(xx) symbol_function(datum_symbol(xx))

/*
 * Disassemble instructions from l to n
 */
Export void disasm(aw, s, e)
AwlWidget aw;
int s, e;
{
     int nargs, insn, i;
     char *op;

     if (e > awl_etext(aw))
	  e = awl_etext(aw);
     else if (s == e)
	  e += 1;
     while (s < e) {
	  /* print any labels set to this address */
	  hash_traverse(awl_top(aw), show_labels, &s);

	  fprintf(awl_err(aw), "%#08x: ", s);
	  if ((nargs = op_nargs(insn = datum_int(awl_text(aw)[s]))) >= 0) {
	       op = op_name(datum_int(awl_text(aw)[s++]));
	       fputs(op, awl_err(aw));
	       for (i = strlen(op); i < 8; i++)
		    fputc(' ', awl_err(aw));
	       if (nargs == 0)
		    fputc('\n', awl_err(aw));
	       else if (nargs == 1) {
		    switch(insn) {
		    case ALU:
			 fprintf(awl_err(aw), "%s\n",
				 alu_name(datum_int(awl_text(aw)[s++])));
			 break;

		    case JUMP:
		    case JUMPZ:
			 fprintf(awl_err(aw), "%s\n",
				 symbol_name(datum_symbol(awl_text(aw)[s++])));
			 break;

		    case CALL:
			 if (symbol_type(datum_symbol(awl_text(aw)[s])) ==
			     FUNC_LAYOUT)
			      fprintf(awl_err(aw), "WL:<%s> ",
				      symbol_name(datum_symbol(awl_text(aw)[s])));
			 else
			      fprintf(awl_err(aw), "<%s> ",
				      symbol_name(datum_symbol(awl_text(aw)[s])));
			 if (function_uaddr(SFDS(awl_text(aw)[s])) ==
			     UNDEF_ADDR) {
			      fprintf(awl_err(aw), "[undef]\n");
			      s++;
			 }
			 else {
			      fprintf(awl_err(aw), "[%#0x]\n",
				      function_addr(SFDS(awl_text(aw)[s])));
			      s++;
			 }
			 break;

		    case OPEN:
			 fprintf(awl_err(aw), "%s\n",
				map_wclass_to_wname(datum_wclass(awl_text(aw)[s++])));
			 break;

		    case POP:
			 switch(datum_int(awl_text(aw)[s++])) {
			 case NODUP:
			      fprintf(awl_err(aw), "[normal store]\n");
			      break;

			 case DISCARD:
			      fprintf(awl_err(aw), "[discard]\n");
			      break;

			 case DUPVAL:
			      fprintf(awl_err(aw), "[store and dup value]\n");
			      break;

			 case DUPTARG:
			      fprintf(awl_err(aw), "[store and dup target]\n");
			      break;

			 default:
			      fprintf(awl_err(aw), "[?? bad pop ??]\n");
			      break;
			 }
			 break;

		    case RTS:
			 fprintf(awl_err(aw), "%s\n",
				 datum_int(awl_text(aw)[s++]) ? "[with value]"
				 : "[without value]");
			 break;

		    default:
			 fprintf(awl_err(aw), "???\n");
			 break;
		    }
	       }
	       else if (nargs == 2) {	/* PUSH */
		    char *pstr;
		    int cnst = FALSE;
		    int type = datum_int(awl_text(aw)[s]);
		    Datum data = awl_text(aw)[s + 1];

		    if (type != 0) {
			 if (type < 0) {
			      cnst = TRUE;
			      type = -type;
			 }
			 if (type == STRING)
			      pstr = "\"%s\"\n";
			 else
			      pstr = "%s\n";
			 if (cnst)
			      fprintf(awl_err(aw), "<C:%s> ", type_name(type));
			 else
			      fprintf(awl_err(aw), "<%s> ", type_name(type));
			 if (type == CHAR)
			      fprintf(awl_err(aw), "\'%#0x\'\n", datum_char(data));
			 else if (type == SYMBOL)
			      fprintf(awl_err(aw), "<%s>\n",
				      symbol_printname(datum_symbol(data)));
			 else
			      fprintf(awl_err(aw), pstr,
				      datum_string_value(aw, type, data));
		    }
		    else
			 fprintf(awl_err(aw), "*ILLEGAL PUSH ARGS*\n");
		    s += 2;
	       }
	       else
		    fprintf(awl_err(aw), "???\n");
	  }
	  else {
	       fprintf(awl_err(aw), "???\n");
	       s++;
	  }
     }
}
