#ifndef lint
static char *rcsid = "$Header: /usr/src/local/awl/RCS/eval.c,v 2.7 90/04/19 20:05:02 jkh Exp $";
#endif

/*
 *
 *                   Copyright 1989, 1990
 *                    Jordan K. Hubbard
 *
 *		  PCS Computer Systeme, GmbH.
 *	             Munich, West Germany
 *
 *
 * This file is part of AWL.
 *
 * AWL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * AWL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AWL; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

/*
 * This file contains the guts of awl. Guts is actually a pretty
 * descriptive word since looking at this code will probably induce
 * a reaction similar to that experienced by most people when looking
 * at salmon or deer vicera. What can I say. Brute-force interpreters
 * tend to be somewhat ugly, no matter what one does. Just grit your teeth.
 *
 * $Log:	eval.c,v $
 * Revision 2.7  90/04/19  20:05:02  jkh
 * Alpha checkin.
 * 
 * Revision 2.6  90/03/30  22:00:59  jkh
 * Changed the way the opcode and alu tables are declared.
 * 
 * Revision 2.5  90/03/30  01:07:17  jkh
 * Added do_optional_jump() routine.
 * 
 * Revision 2.4  90/03/29  18:28:20  jkh
 * Fixed bugs with ASOC.
 * 
 * Revision 2.3  90/03/26  12:41:38  jkh
 * Fixed broken ASOC.
 * 
 * Revision 2.2  90/03/26  03:40:49  jkh
 * Another typo fixed.
 * 
 * Revision 2.0  90/03/26  01:49:31  jkh
 * pre-beta check-in
 * 
 */

#include "AwlP.h"
#include "y.tab.h"

Export Datum TrueDatum;		/* constant for TRUE */
Export Datum FalseDatum;	/* constant for FALSE */
Export Datum PopDupValue;	/* tell POP to duplicate the value */
Export Datum PopDupTarget;	/* tell POP to duplicate the target */
Export Datum PopDupNone;	/* tell POP not to duplicate anything */
Export Datum PopDiscard;	/* tell POP to simply discard a value */

Local void call(), alu(), open_w(), close_w(), ares_w();
Local void fp_exception(), zero_divide();
Local int asoc_value();
Local Resource res_link();

Local AwlWidget Executor;	/* for signals */
Import void add_builtins();
Import void add_predefined_symbols();

/*
 * Our hypothetical "stack machine"'s instruction set contains the following
 * OPs. A little FORTH-like, really. The opcodes are defined as tokens in
 * y.tab.h. All stack values (and variables) are typed.
 *
 * General instructions:
 *
 * ALU   <opcode>	does an ALU op (see below), coercing data apropriately.
 * ARES			create resource named [pop] with value [pop].
 * CALL	 <symbol>	calls function ref'd by <symbol>.
 * CLOS			closes an open widget declaration.
 * DUP			duplicate top of stack.
 * HALT			stop evaluation (for debugging).
 * JUMP	 <addr>		jump to <addr>.
 * JUMPZ <addr>		jump to <addr> if [pop] == 0.
 * LINK			sets up of a new parameter call frame.
 * OPEN	 		opens a widget declaration of class <wclass>.
 * POP	 <poptype>	pops one or two stack elements depending on poptype.
 * PUSH  <type> <value>	pushes <value> of <type> onto stack.
 * RTS	 <val?>		return from a function, conditionally with value.
 *
 * ALU opcodes:
 *
 * ADDR			push &[pop].
 * AND			push [pop] & [pop].
 * ASOC			assoc reference [pop].
 * CAST			push convert [pop] to [pop].
 * COMP			push ~[pop] (unary).
 * COPY			push (copy of)[pop].
 * DIV			push [pop] / [pop].
 * DOT			push [pop] submember [pop].
 * EE			push [pop] == [pop].
 * EQ			push [pop] = [pop].
 * GE			push [pop] >= [pop].
 * GT			push [pop] > [pop].
 * LAND			push [pop] && [pop].
 * LE			push [pop] <= [pop].
 * LOR			push [pop] || [pop].
 * LT			push [pop] < [pop].
 * MBR			push [pop](value) ?? [pop](array var).
 * MINUS		push -[pop].
 * MOD			push [pop] % [pop].
 * MULT			push [pop] / [pop].
 * NE			push [pop] != [pop].
 * NOT			push ![pop].
 * NUKE			!! ![pop].
 * OR			push [pop] | [pop].
 * PLUS			push [pop] + [pop].
 * SL			push [pop] << [pop].
 * SR			push [pop] >> [pop].
 * SUB			push [pop] - [pop].
 * TRI			push [pop] ? [pop] : [pop].
 * XOR			push [pop] ^ [pop].
 *
 */

/*
 * The instruction table. The order may seem peculiar but it makes a certain
 * warped amount of sense. The most frequently used insns are grouped together
 * by number of arguments, followed by the next less-frequently used groups,
 * etc. This lets us find instructions in a maximum of 2 compares (but usually
 * only one) later. If you're going to change this, be sure to look at awl.y
 * and op_nargs() in evalutils.c first since these need to stay in sync.
 */
Export InsnTable opcode_table[] = {
     { ALU,	1,	"ALU"   },
     { CALL,	1,	"CALL"  },
     { JUMP,	1,	"JUMP"  },
     { JUMPZ,	1,	"JUMPZ" },
     { POP,	1,	"POP"   },
     { RTS,	1,	"RTS"   },
     { ARES,	0,	"ARES"  },
     { CLOS,	0,	"CLOS"  },
     { DUP,	0,	"DUP"   },
     { HALT,	0,	"HALT"	},
     { LINK,	0,	"LINK"	},
     { OPEN,	0,	"OPEN"  },
     { PUSH,	2,	"PUSH"  },
};

/*
 * Declared in order of increasing precendence. Don't change this unless
 * you know what you're doing!
 */
Export InsnTable alu_opcodes[] = {
     { EQ,	2,	"=",	},
     { LOR,	2,	"||"	},
     { LAND,	2,	"&&"	},
     { OR,	2,	"|"	},
     { XOR,	2,	"^"	},
     { AND,	2,	"&"	},
     { COPY,	1,	"<>"	},
     { NUKE,	1,	"!!"	},
     { EE,	2,	"=="	},
     { NE,	2,	"!="	},
     { LT,	2,	"<"	},
     { LE,	2,	"<="	},
     { GT,	2,	">"	},
     { GE,	2,	">="	},
     { SL,	2,	"<<"	},
     { SR,	2,	">>"	},
     { PLUS,	2,	"+"	},
     { MINUS,	1,	"-(neg)"},
     { MULT,	2,	"*"	},
     { DIV,	2,	"/"	},
     { MOD,	2,	"%"	},
     { NOT,	1,	"!"	},
     { ADDR,	1,	"&(adr)"},
     { COMP,	1,	"~"	},
     { CAST,	2,	"()"	},
     { MBR,	2,	"??"	},
     { DOT,	2,	"."	},
     { ASOC,	2,	"[]"	},
     { SUB,	2,	"-"	},
     { TRI,	3,	"?:"	},
     { _LASTALU,0,	NULL	},
};

Export int opcode_table_size = sizeof(opcode_table) / sizeof(InsnTable);
Export int alu_table_size = sizeof(alu_opcodes) / sizeof(InsnTable);

/* Initialize text/stack segments. */
Export void do_init(aw)
AwlWidget aw;
{
     awl_estack(aw) = STACK_INCR;
     awl_etext(aw) = TEXT_INCR;

     awl_stack(aw) = (Value *)XtMalloc(STACK_INCR * sizeof(Value));
     awl_text(aw) = (Datum *)XtMalloc(TEXT_INCR * sizeof(Datum));

     awl_sp(aw) = awl_oldsp(aw) = awl_fp(aw) = awl_pc(aw) = 0;
     awl_nparms(aw) = 0;
     awl_lastinsn(aw) = 0;
     awl_lastpc(aw) = 1;

     awl_in(aw) = stdin;
     awl_out(aw) = stdout;
     awl_err(aw) = stderr;
     awl_radix(aw) = 10;
     awl_sep(aw) = XtNewString("\n");
     awl_ctable(aw) = NULL;

     awl_top(aw) = hash_create(GLOBAL_HASH_SIZE);

     datum_int(*(awl_text(aw))) = HALT;

     /* Initialize various constants */
     datum_int(TrueDatum) = TRUE;
     datum_int(FalseDatum) = FALSE;
     datum_int(PopDupValue) = DUPVAL;
     datum_int(PopDupTarget) = DUPTARG;
     datum_int(PopDupNone) = NODUP;
     datum_int(PopDiscard) = DISCARD;

     add_builtins(aw);
     add_predefined_symbols(aw);

#if defined(SIGFPE)
     signal(SIGFPE, fp_exception);
#endif
#if defined(SIGZERO)
     signal(SIGZERO, zero_divide);
#endif

     if (awl_lineinfo(aw)) {
	  awl_ltabsize(aw) = LTAB_INCR;
	  awl_ltabcurr(aw) = 0;
	  awl_ltab(aw) = (LTabEnt *)XtMalloc(LTAB_INCR * ltab_size);
	  ltab_line(awl_ltab(aw)[0]) = ltab_start(awl_ltab(aw)[0]) =
	       ltab_last(awl_ltab(aw)[0]) = 0;
	  ltab_file(awl_ltab(aw)[0]) = awl_file(aw);
     }
     else
	  awl_ltab(aw) = (LTabEnt *)0;

     awl_bkpts(aw) = NULL;

     if (awl_version(aw)) {
	  printf("AWL Version: %d.%d, compiled for the %s widget set.\n",
		 VERSION, PATCHLEVEL,
#ifdef ATHENA
		 "Athena"
#elif MOTIF
		 "Motif"
#elif XW
		 "Xw (HP)"
#else
		 "(unknown???)"
#endif
		 );
     }
}

/* buzz through and register any external information the user passed */
Export void do_externs(aw, el)
AwlWidget aw;
AwlExternList el;
{
     register int i, type, nargs;
     register String name;

     for (i = 0; name = el[i].name; i++) {
	  type = 0;

	  switch (el[i].type) {
	  case TYPE_CHAR:
	       type = CADDR;
	       break;

	  case TYPE_INT:
	       type = IADDR;
	       break;

	  case TYPE_FLOAT:
	       type = FADDR;
	       break;

	  case TYPE_ANY:
	       type = ANY;

	  case TYPE_STRING:
	       type = STRING;
	       break;

	  case TYPE_LIST:
	       type = LIST;
	       break;

	  case TYPE_FUNCTION:
	       type = FUNC_EXTERN;
	       nargs = -1;
	       break;

	  case TYPE_A_FUNCTION:
	       type = FUNC_ACTION;
	       nargs = -1;
	       break;

	  case TYPE_C_FUNCTION:
	       type = FUNC_CALLBACK;
	       nargs = 1;
	       break;
	  }
	  if (type) {
	       if (is_function_type(type)) {
		    Address a;
		    Symbol tmp;

		    addr_ep(a) = (GenFunc)el[i].ptr;
		    tmp = symbol_def_function(aw, name, type, nargs, a, TRUE);
		    function_data(symbol_function(tmp)) = el[i].client_data;
	       }
	       else {
		    Value val;
		    Symbol sym;

		    if (is_allocated_type(type))
			 value_set(val, STATIC, type, aobj,
				   new_aobj(el[i].ptr));
		    else
			 value_set(val, DATA, type, any, (Generic)el[i].ptr);
		    sym = symbol_new(aw, name, type);
		    symbol_value(sym) = val;
		    symbol_add(sym);
	       }
	  }
	  else if (el[i].type == TYPE_WCLASS) {
	       Datum d1;

	       datum_wclass(d1) = (WidgetClass)el[i].ptr;
	       symbol_def_constant(aw, name, WCLASS, d1);
	  }
	  else
	       exec_error(aw, "Illegal type #%d for external '%s'",
			  el[i].type, name);
     }
}
     
/* check an address for validity */
#define CHECK_TEXT_ADDR(aw, addr) \
     if (addr == UNDEF_ADDR) \
     	 exec_error(aw, "Jump to undefined label attempted"); \
     else if (addr < 0 || addr > awl_etext(aw)) \
         exec_error(aw, "Illegal TEXT address '%x'", addr)

/*
 * Compile an instruction into the AWL "text segment".
 */
Export void do_emit(aw, insn, arg1, arg2)
AwlWidget aw;
int insn;
Datum arg1, arg2;
{
     int nargs;

     if ((nargs = op_nargs(insn)) < 0) {	/* instruction not found */
	  yyerror("emit: Invalid instruction at addr 0x%08x\n",
		  awl_lastpc(aw));
	  return;
     }
     else if (nargs > 2) {			/* illegal number of args */
	  yyerror("emit: Instruction %d has > 2 args!\n",
		  insn);
	  return;
     }
     else if (datum_int(awl_text(aw)[awl_lastinsn(aw)]) == insn && insn == RTS)
	  return;	/* don't do redundant RTS */

     adjust(aw, awl_text(aw), nargs + 1);
     if (awl_ltab(aw)) {
	  Local String file = NULL, check = NULL;

	  if (check != yy_file) {
	       check = yy_file;
	       file = XtNewString(check);
	  }
	  if (ltab_line(awl_ltab(aw)[awl_ltabcurr(aw)]) == (yy_line - 1))
	       ltab_last(awl_ltab(aw)[awl_ltabcurr(aw)]) = awl_lastpc(aw)
		    + nargs;
	  else {
	       adjust(aw, awl_ltab(aw), 1);
	       awl_ltabcurr(aw)++;
	       ltab_line(awl_ltab(aw)[awl_ltabcurr(aw)]) =
		    (yy_line) ? yy_line - 1 : 0;
	       ltab_file(awl_ltab(aw)[awl_ltabcurr(aw)]) = file;
	       ltab_start(awl_ltab(aw)[awl_ltabcurr(aw)]) = awl_lastpc(aw);
	       ltab_last(awl_ltab(aw)[awl_ltabcurr(aw)]) = awl_lastpc(aw)
		    + nargs;
	  }
     }
     awl_lastinsn(aw) = awl_lastpc(aw);
     datum_int(awl_text(aw)[awl_lastpc(aw)++]) = insn;
     if (nargs >= 1)
	  awl_text(aw)[awl_lastpc(aw)++] = arg1;
     if (nargs == 2)
	  awl_text(aw)[awl_lastpc(aw)++] = arg2;
}

/* shorthand for ALU ops */
Export Inline void do_alu(aw, op)
AwlWidget aw;
int op;
{
     Datum d1;

     datum_int(d1) = op;
     do_emit(aw, ALU, d1);
}

/* shorthand for pushes */
Export Inline void do_push(aw, type, data)
AwlWidget aw;
int type;
Datum data;
{
     Datum d1;

     if (type != SYMBOL)
	  datum_int(d1) = -type;
     else
	  datum_int(d1) = SYMBOL;
     do_emit(aw, PUSH, d1, data);
}

/*
 * This is the main routine for storing values in AWL. It handles
 * storing directly to symbols and/or areas of the stack. It also ensures
 * that malloc'd values being overwritten are properly free'd and that
 * source values are properly duplicated.
 */
Export void do_store(aw, valp, val)
register AwlWidget aw;
register Value *valp;
Value val;
{
     Resource r;
     Symbol sym;
     Arg a[1];
     int i;

     if (value_type(*valp) == SYMBOL) {
	  if (value_class(*valp) == FUNCTION) {
	       exec_warn(aw, "Illegal function value for LHS, ignored.");
	       return;
	  }
	  sym = value_symbol(*valp);
	  if (symbol_type(sym) == _FP)
	       valp = awl_stack(aw) + FRAME_OFFSET(symbol_int(sym));
	  else if (symbol_type(sym) == _SP)
	       valp = awl_stack(aw) + (awl_sp(aw) + symbol_int(sym));
	  else
	       valp = symbol_addr(sym);
     }
     if (value_class(*valp) == ALLOCATED) {
	  if (value_type(*valp) == STRING && value_type(val) == STRING)
	       strcpy(value_string(*valp), value_string(val));
	  else
	       exec_error(aw, "Attempt to change type of declare'd variable.");
     }
     switch (value_type(*valp)) {
     case RESOURCE:
	  i = 0;
	  r = (Resource)value_any(*valp);
	  if (value_type(res_value(r)) == LAYOUT) {
	       res_value(r) = val;
	       if (wl_id(res_parent(r))) {
	            wset_arg_by_value(a, &i, res_name(r), val);
	            XtSetValues(wl_id(res_parent(r)), a, i);
	       }
	  }
	  else if (value_type(res_value(r)) == WIDGET) {
	       wset_arg_by_value(a, &i, res_name(r), val);
	       XtSetValues(value_widget(res_value(r)), a, i);
	       XtFree(r);
	  }
	  else
	       exec_error(aw, "Bad resource entry for type %s",
			  type_name(value_type(res_value(r))));
	  break;

     case CADDR:
	  ((char *)value_any(*valp))[0] = value_char(coerce(aw, val, CHAR));
	  break;

     case SADDR:
	  XtFree(((String *)value_any(*valp))[0]);
	  ((String *)value_any(*valp))[0] = 
	       value_string(coerce(aw, val, STRING));
	  break;

     case IADDR:
	  ((int *)value_any(*valp))[0] = value_int(coerce(aw, val, INT));
	  break;

     case FADDR:
	  ((FTYPE *)value_any(*valp))[0] = value_float(coerce(aw, val, FLOAT));
	  break;

     default:
	  do_free(valp);
	  *valp = val;
	  break;
     }
}

/*
 * These routines handle all loading of data from various parts of
 * the AWL environment (stack, symbol table, direct).
 */

/* Handle register loads */
Export Inline Value do_register_load(aw, val)
AwlWidget aw;
Value val;
{
     if (value_type(val) == _FP)
	  val = awl_stack(aw)[FRAME_OFFSET(value_int(val))];
     else if (value_type(val) == _SP)
	  val = awl_stack(aw)[awl_sp(aw) + value_int(val)];
     else
	  exec_warn(aw, "Illegal register value for load");
     return(val);
}
    
Export Value do_load(aw, val, bump)
AwlWidget aw;
Value val;
Boolean bump;
{
     Value ret;
     Resource r;
     Arg a[1];
     String tmp;
     int typ;

     if (value_type(val) == SYMBOL
	 && symbol_class(value_symbol(val)) != FUNCTION) {
	  val = symbol_value(value_symbol(val));
	  if (is_register(value_type(val)))
	       val = do_register_load(aw, val);
	  /*
	   * The symbol's contents may also be indirect. If so, call ourselves
	   * recursively to satisfy the additional indirection.
	   */
	  if (value_type(val) == SYMBOL)
	       val = do_load(aw, val, bump);
     }
     switch(value_type(val)) {
     case RESOURCE:
	  r = (Resource)value_any(val);
	  /* If we don't already know it, look it up */
	  typ = value_type(res_value(r));
	  if (typ == LAYOUT || typ == WIDGET) {
	       value_type(res_value(r)) = ANY;
	       XtSetArg(a[0], res_name(r), &value_any(res_value(r)));
	       if (typ == LAYOUT) {
	            XtGetValues(wl_id(res_parent(r)), a, 1);
	       }
	       else {
		    XtGetValues(value_widget(res_value(r)), a, 1);
	       }
	       ret = res_value(r);
	       if (typ == WIDGET) { /* was only temporary */
		    XtFree(r);
	       }
	  }
	  else
	       exec_error(aw, "Bad type %s for RESOURCE", type_name(typ));
	  break;

     case CADDR:
	  value_set(ret, DATA, CHAR, char, ((char *)value_any(val))[0]);
	  break;

     case IADDR:
	  value_set(ret, DATA, INT, int, ((int *)value_any(val))[0]);
	  break;

     case FADDR:
	  value_set(ret, DATA, FLOAT, float, ((FTYPE *)value_any(val))[0]);
	  break;

     case SADDR:
	  tmp = ((String *)value_any(val))[0];
	  value_set(ret, STATIC, STRING, aobj, new_aobj(tmp));
	  break;

     default:
	  ret = val;
     }
     if (is_allocated_type(value_type(ret))
	 && !is_precious(value_class(ret))
	 && bump)
	  ++aobj_cnt(value_aobj(ret));
     return(ret);
}

/* keep the stack clean by eliminating unused RHS values */
Export Inline void do_discard(aw)
AwlWidget aw;
{
     /*
      * If the last insn was a duplicating POP, simply change
      * it to a non-duplicating POP, otherwise emit a discarding POP. 
      */
     if (datum_int(awl_text(aw)[awl_lastinsn(aw)]) == POP)
	  awl_text(aw)[awl_lastinsn(aw) + 1] = PopDupNone;
     else
	  do_emit(aw, POP, PopDiscard);
}

/*
 * In some cases, we emit a JUMP to save ourselves from some horrible
 * fate that would occur by simply "falling through" to another section
 * of code. These JUMPs are redundant if immediately preceeded by another
 * JUMP, so conditionally don't emit them.
 */
Export Inline void do_optional_jump(aw, addr)
AwlWidget aw;
Symbol addr;
{
     /* Only do it if necessary */
     if (datum_int(awl_text(aw)[awl_lastinsn(aw)]) != JUMP) {
	  Datum d1;

	  datum_symbol(d1) = addr;
	  do_emit(aw, JUMP, d1);
     }
}

/* Push the stack around */
Export Inline void push(aw, v)
AwlWidget aw;
Value v;
{
     adjust(aw, awl_stack(aw), 1);
     awl_stack(aw)[awl_sp(aw)++] = v;
}

/* Rip its head off */
Export Inline Value pop(aw)
AwlWidget aw;
{
     if (awl_sp(aw) == 0)
	  exec_error(aw, "Pop: Stack underflow");
     return(awl_stack(aw)[--awl_sp(aw)]);
}

/*
 * Peek down its neckline (really, if you can't figure these out,
 * you should go home). If SP == 0 the return value is bogus.
 * It's up to the caller to check for that before calling this routine.
 */
Export Inline Value tos(aw)
AwlWidget aw;
{
     return(awl_stack(aw)[awl_sp(aw) - 1]);
}

/*
 * This is the main evaluation loop.
 */
Export Value Eval(name)
Symbol name;
{
     AwlWidget aw = symbol_root(name);
     Value val;
     Symbol sym;
     int addr, vt, class;

     debug(aw, "Evaluating '%s'. %x items on stack. Code segment size: %x\n",
	   symbol_name(name), awl_sp(aw), awl_lastpc(aw));

     /*
      * If call() is to a user defined function, all it will do is set the
      * PC to the beginning of it, in which case we can
      * fall through and start interpreting the opcodes. If it was a callback
      * or builtin, then call() has already done all we want to do and PC will
      * still be zero.
      */
     call(name);

     while (awl_pc(aw)) {
	  /* switch on the current instruction */
	  if (awl_trace(aw))
	       disasm(aw, awl_pc(aw), awl_pc(aw) +
		      op_nargs(awl_text(aw)[awl_pc(aw)]));
	  if (awl_bkpts(aw)) {
	       Breakpoint tmp = awl_bkpts(aw);

	       while (tmp) {
		    if (bkpt_enable(tmp) && bkpt_addr(tmp) == awl_pc(aw)) {
			 /* If bkpt func returns TRUE, abort */
			 if (bkpt_func(tmp)(aw, awl_pc(aw))) {
			      Value arg;

			      value_clear(arg);
			      return(arg);
			 }
			 else
			      break;
		    }
		    tmp = bkpt_next(tmp);
	       }
	  }

	  switch(datum_int(awl_text(aw)[awl_pc(aw)])) {
	  case ALU:
	       /* let the alu rtn handle it */
	       alu(aw, datum_int(awl_text(aw)[awl_pc(aw) + 1]));
	       awl_pc(aw) += 2;
	       break;

	  case CALL:
	       call(datum_symbol(awl_text(aw)[++awl_pc(aw)]));
	       break;

	  case DUP:
	       if (awl_sp(aw))
		    push(aw, tos(aw));
	       else
		    exec_error(aw, "DUP: stack empty");
	       ++awl_pc(aw);
	       break;

	  case LINK:
	       do_link(aw);
	       ++awl_pc(aw);
	       break;

	  case HALT:
	       awl_pc(aw) = 0;
	       break;

	  case JUMP:
	       sym = datum_symbol(awl_text(aw)[awl_pc(aw) + 1]);
	       if (symbol_class(sym) != LABEL)
		    exec_error(aw, "JUMP to non-label attempted.");
	       else {
		    addr = symbol_int(sym);
		    CHECK_TEXT_ADDR(aw, addr);
		    awl_pc(aw) = addr;
	       }
	       break;

	  case JUMPZ:
	       val = coerce(aw, pop(aw), ANY);
	       sym = datum_symbol(awl_text(aw)[awl_pc(aw) + 1]);
	       if (symbol_class(sym) != LABEL)
		    exec_error(aw, "JUMPZ to non-label attempted.");
	       else {
		    Boolean cond;

		    addr = symbol_int(sym);
		    CHECK_TEXT_ADDR(aw, addr);
		    if (is_allocated_type(value_type(val)))
			 cond = (aobj_ptr(value_aobj(val)) == NULL);
		    else
			 cond = (value_any(val) == NULL);
		    if (cond)
			 awl_pc(aw) = addr;		/* take the jump */
		    else
			 awl_pc(aw) += 2;		/* just move over it */
	       }
	       break;

	  case PUSH:
	       vt = datum_int(awl_text(aw)[awl_pc(aw) + 1]);
	       if (vt < 0) {
		    class = CONSTANT;
		    vt = -vt;
	       }
	       else
		    class = DATA;
	       value_type(val) = vt;
	       /* LAYOUT is a special case kludge for layouts */
	       if (vt == LAYOUT)
		    value_set(val, class, vt, any, (Generic)awl_active(aw));
	       /* special case float in case a special insn is needed */
	       else if (vt == FLOAT)
		    value_set(val, class, vt, float,
			      datum_float(awl_text(aw)[awl_pc(aw) + 2]));
	       else if (is_allocated_type(vt))
		    value_set(val, class, vt, aobj,
			      datum_aobj(awl_text(aw)[awl_pc(aw) + 2]));
	       else {
		    value_any(val) = datum_any(awl_text(aw)[awl_pc(aw) + 2]);
		    value_class(val) = class;
	       }
	       push(aw, val);
	       awl_pc(aw) += 3;
	       break;

	  case RTS:
	       /* save return value if necessary. */
	       if (datum_int(awl_text(aw)[awl_pc(aw) + 1]) == TRUE)
		    val = coerce(aw, pop(aw), ANY);
	       else
		    value_clear(val);

	       do_rts(aw);
	       /*
		* Push the function's value. Functions that don't return
		* a value get a default value of (int)0 (see above).
		* Since unused values are always popped, this will simply
		* be discarded if (typically) ignored.
		*/
	       push(aw, val);
	       break;

	  case POP:
	       /* is it a discarding pop? */
	       if (datum_int(awl_text(aw)[awl_pc(aw) + 1]) == DISCARD) {
		    Value val;

		    val = pop(aw);
		    do_free(&val);
	       }
	       else {
		    Value var;

		    val = do_load(aw, pop(aw), TRUE);
		    var = pop(aw); /* just pop since we want raw target */
		    do_store(aw, &var, val);
		    if (datum_int(awl_text(aw)[awl_pc(aw) + 1]) == DUPVAL)
			 push(aw, val);
		    else if (datum_int(awl_text(aw)[awl_pc(aw)+1]) == DUPTARG)
			 push(aw, var);
	       }
	       awl_pc(aw) += 2;
	       break;

	  case ARES:
	       ares_w(aw);
	       ++awl_pc(aw);
	       break;

	  case CLOS:
	       close_w(aw);
	       ++awl_pc(aw);
	       break;

	  case OPEN:
	       open_w(aw);
	       ++awl_pc(aw);
	       break;

	  default:
	       exec_error(aw, "Unknown instruction");
	       break;
	  }
     }
     /* we're at the top level, return the top of stack */
     if (awl_sp(aw))
	  val = coerce(aw, pop(aw), ANY);
     else {
	  value_int(val) = 0;
	  value_type(val) = INT;
     }
     return(val);
}

/* do a LINK */
Export Inline void do_link(aw)
AwlWidget aw;
{
     Value val;

     value_set(val, DATA, INT, int, awl_oldsp(aw));
     awl_oldsp(aw) = awl_sp(aw);
     push(aw, val);
}

/* do an RTS */
Export Inline void do_rts(aw)
AwlWidget aw;
{
     int i, nlocals;

     /*
      * Local variables and parameters may have accessed globally
      * allocated values. Make sure we unreference these as necessary.
      */
     nlocals = (awl_sp(aw) - awl_fp(aw)) - 1;
     debug(aw, "do_rts: Popping off %d parameters and %d locals",
	   awl_nparms(aw), nlocals);

     /* first, the parameters */
     for (i = 1; i <= awl_nparms(aw); i++)
	  if (is_allocated_type(value_type(awl_stack(aw)[FRAME_OFFSET(-i)])))
	       --aobj_cnt(value_aobj(awl_stack(aw)[FRAME_OFFSET(-i)]));

     /* now the locals */
     for (i = 1; i <= nlocals; i++)
	  if (is_allocated_type(value_type(awl_stack(aw)[FRAME_OFFSET(i)])))
	       --aobj_cnt(value_aobj(awl_stack(aw)[FRAME_OFFSET(i)]));

     awl_pc(aw) = value_int(awl_stack(aw)[awl_fp(aw) - 2]);
     awl_nparms(aw) = value_int(awl_stack(aw)[awl_fp(aw) - 1]);
     awl_fp(aw) = value_int(awl_stack(aw)[awl_fp(aw)]);
     awl_sp(aw) = awl_oldsp(aw);
     awl_oldsp(aw) = value_int(awl_stack(aw)[awl_sp(aw)]);
}

Local String *parms_as_list(aw)
AwlWidget aw;
{
     register int i;
     String *ret;
     Value val;

     ret = (String *)XtMalloc((awl_nparms(aw) + 1) * sizeof(String *));
     for (i = 1; i <= awl_nparms(aw); i++) {
	  val = get_arg(aw, i, STRING, FALSE);
	  ret[i - 1] = value_string(val);
     }
     ret[i - 1] = NULL;
     return(ret);
}

/* call a function */
Local void call(sym)
Symbol sym;
{
     AwlWidget aw = symbol_root(sym);
     register int addr, i, nparms;
     Function f;
     Value val;

     /* is it a function pointer rather than a function? */
     if (symbol_type(sym) == SYMBOL)
	  sym = symbol_symbol(sym);
     if (!symbol_is_function(sym)) {
	  exec_warn(aw, "Target of CALL is not a function");
	  ++awl_pc(aw);
	  return;
     }
     else
	  f = symbol_function(sym);


     /* How many parameters did we get? */
     nparms = awl_sp(aw) ? (awl_sp(aw) - awl_oldsp(aw) - 1) : 0;

     /* Check to see if we've been called with too few parameters. */
     if (function_nparms(f) > 0 && nparms < function_nparms(f)) {
	  int reqd;

	  value_clear(val);
	  reqd = function_nparms(f) - nparms;
	  nparms = function_nparms(f);
	  debug(aw, "Padding %d missing parameter(s) to '%s' call",
		reqd, symbol_name(sym));
	  for (i = 1; i <= reqd; i++)
	       push(aw, val);
     }
     /*
      * Since all variables are passed by value, make sure all variables
      * passed on the stack as parameters are dereferenced. This includes
      * bumping up reference counters on allocated objects, of course.
      */
     for (i = nparms; i > 0; i--) {
	  if (value_type(awl_stack(aw)[awl_sp(aw) - i]) == SYMBOL)
	       awl_stack(aw)[awl_sp(aw) - i] =
		    do_load(aw, awl_stack(aw)[awl_sp(aw) - i], TRUE);
     }

     /* Now start constructing the new frame */
     value_set(val, DATA, INT, int, 0);
     value_int(val) = awl_pc(aw) ? awl_pc(aw) + 1 : 0;
     push(aw, val);     /* return PC */
     value_int(val) = awl_nparms(aw);
     awl_nparms(aw) = nparms;
     push(aw, val);	/* saved nparms */
     /* Move FP up and leave it pointing at its old value */
     value_int(val) = awl_fp(aw);
     awl_fp(aw) = awl_sp(aw);
     push(aw, val);	/* old FP (push is post-increment) */

     /* Now make space for local variables, if any */
     value_clear(val);
     for (i = 0; i < function_nlocals(f); i++)
	  push(aw, val);

     switch(symbol_type(sym)) {

     case FUNC_UNDEF:
	  exec_error(aw, "function '%s' was never defined", symbol_name(sym));
	  ++awl_pc(aw);
	  break;

     case FUNC_USER:
     case FUNC_LAYOUT:
	  addr = function_uaddr(f);
	  CHECK_TEXT_ADDR(aw, addr);		/* bounds check */
	  awl_pc(aw) = addr;
	  break;

     case FUNC_BUILTIN:
	  val = (*function_faddr(f))(aw);
	  do_rts(aw);
	  /* push saved value after frame restored ("val" can be garbage) */
	  push(aw, val);
	  break;

     case FUNC_EXTERN:
	  value_set(val, DATA, ANY, any,
		    (*function_eaddr(f))(aw, function_data(f), awl_nparms(aw),
					 parms_as_list(aw)));
	  do_rts(aw);
	  push(aw, val);
	  break;

     case FUNC_CALLBACK:
	  if (awl_nparms(aw) > 1)
	       exec_warn(aw, "Extra parameters to CALLBACK function ignored");
	  val = get_arg(aw, 1, ANY, FALSE);
	  (*function_faddr(f))(aw, function_data(f), value_any(val));
	  do_rts(aw);
	  value_clear(val);
	  push(aw, val);
	  break;

     case FUNC_ACTION:
	  (*function_eaddr(f))(aw, NULL, parms_as_list(aw), &awl_nparms(aw));
	  do_rts(aw);
	  break;

     default:
	  fprintf(stderr, "Unknown function type for '%s'!\n",
		  symbol_name(sym));
	  ++awl_pc(aw);
	  break;
     }
}

/*
 * Various handy but confusing macros for doing ALU ops.
 */

#define OP_DECL \
  Value a0, a1, a2, res

#define POP2(aw) \
  a2 = coerce(aw, pop(aw), ANY); \
  a1 = coerce(aw, pop(aw), ANY);

/*
 * The CHAR and INT ops can probably be combined, but I'm being careful. God
 * only knows what weird 18 bit int machines might be lurking out there.
 */
#define ARITH_OP(aw, op, aobj_method) \
  POP2(aw); \
  value_class(res) = value_class(a1); \
  value_type(res) = value_type(a1); \
  switch(value_type(a1)) { \
  case CHAR: \
       value_char(res) = value_char(a1) op value_char(coerce(aw, a2, CHAR)); \
       break; \
  case INT: \
       value_int(res) = value_int(a1) op value_int(coerce(aw, a2, INT)); \
       break; \
  case FLOAT: \
       value_float(res) = value_float(a1) op \
	    value_float(coerce(aw, a2, FLOAT)); \
       break; \
  AOBJCASE: \
       value_aobj(res) = aobj_method(aw, a1, a2); \
       break; \
  default: \
       value_type(res) = INT; \
       value_int(res) = 0; \
       break; \
  }

#define LOGICAL_BINARY_OP(aw, op) \
  POP2(aw); \
  value_class(res) = DATA; \
  value_type(res) = INT; \
  switch(value_type(a1)) { \
  case CHAR: \
       value_int(res) = (value_char(a1) op value_char(coerce(aw, a2, CHAR))); \
       break; \
  case INT: \
       value_int(res) = (value_int(a1) op value_int(coerce(aw, a2, INT))); \
       break; \
  case FLOAT: \
       value_int(res) = (value_float(a1) op \
			 value_float(coerce(aw, a2, FLOAT))); \
       break; \
  case STRING: \
       a0 = coerce(aw, a2, STRING); \
       value_int(res) = (strcomp(value_string(a1), value_string(a0)) op 0); \
       break; \
  default: \
       value_int(res) = (value_any(a1) op value_any(a2)); \
       break; \
  }

#define INTEGER_BINARY_OP(aw, op) \
  a1 = coerce(aw, pop(aw), INT); \
  a2 = coerce(aw, pop(aw), INT); \
  value_set(res, DATA, INT, int, value_int(a1) op value_int(a2));

#define INTEGER_UNARY_OP(aw, op) \
  a1 = coerce(aw, pop(aw), INT); \
  value_set(res, DATA, INT, int, op value_int(a1));

/* do an ALU operation */
Local void alu(aw, op)
AwlWidget aw;
int op;
{
     OP_DECL;
     Value *a3;
     Generic addr;
     int type, idx;
     Symbol sym;
     String tmp;
     Boolean cond;

     /* for the FP error signal routines, in case one goes off */
     Executor = aw;

     switch(op) {

     case ADDR:
	  a1 = pop(aw);
	  if (value_type(a1) == SYMBOL) {
	       sym = value_symbol(a1);

	       if (symbol_type(sym) == _FP)
		    a3 = awl_stack(aw) + FRAME_OFFSET(symbol_int(sym));
	       else if (symbol_type(sym) == _SP)
		    a3 = awl_stack(aw) + awl_sp(aw) + symbol_int(sym);
	       else
		    a3 = &symbol_value(sym);
	  }
	  else
	       a3 = &a1;
	  switch (value_type(*a3)) {
	  case INT:
	       type = IADDR;
	       addr = (Generic)&(value_int(*a3));
	       break;
	  case FLOAT:
	       type = FADDR;
	       addr = (Generic)&(value_float(*a3));
	       break;
	  case CHAR:
	       type = CADDR;
	       addr = &value_char(*a3);
	       break;
	  case STRING:
	       type = SADDR;
	       addr = (Generic)&(value_string(*a3));
	       break;
	  case LIST:
	       type = ANY;
	       addr = (Generic)&(value_list(*a3));
	       break;
	  default:
	       type = ANY;
	       addr = (Generic)&(value_any(*a3));
	       break;
	  }
	  value_set(res, DATA, type, any, addr);
	  break;

     case AND:
	  INTEGER_BINARY_OP(aw, &);
	  break;

     case ASOC:
	  a1 = coerce(aw, pop(aw), ANY);	/* subscript */
	  a2 = pop(aw);				/* symbol/layout */

	  if (value_type(a2) == SYMBOL) {
	       Value a4;
	       Symbol sym;

	       sym = value_symbol(a2);
	       a4 = symbol_value(sym);

	       /* register relative values we don't want */
	       if (symbol_is_register(sym))
		    a4 = do_register_load(aw, a4);

	       if (!asoc_value(aw, a4, a1, &res))
		    value_set(res, DATA, SYMBOL, symbol,
			      symbol_alist_find(sym, a1, TRUE));
	  }
	  else {
	       a2 = coerce(aw, a2, ANY);
	       if (!asoc_value(aw, a2, a1, &res))
		    exec_error(aw, "Illegal array reference");
	  }
	  break;

     case COMP:
	  INTEGER_UNARY_OP(aw, ~);
	  break;

     case COPY:
	  a1 = coerce(aw, pop(aw), ANY);
	  if (is_allocated_type(value_type(a1))) {
	       if (value_type(a1) == STRING) {
		    String tmp = XtNewString(value_string(a1));
		    value_set(res, DATA, STRING, aobj, new_aobj(tmp));
	       }
	       else if (value_type(a1) == LIST) {
		    String *tmp = list_dup(value_list(a1));
		    value_set(res, DATA, LIST, aobj, new_aobj(tmp));
	       }
	       else
		    exec_error(aw, "Unknown allocated type '%s'(%d)",
			       type_name(value_type(a1)), value_type(a1));
	  }
	  else
	       res = a1;
	  break;

     case CAST:
	  a1 = pop(aw);
	  a2 = coerce(aw, pop(aw), ANY);
	  res = do_convert(aw, a1, a2);
	  break;
	  
     case DIV:
          ARITH_OP(aw, /, div_alloc);
	  break;

     case DOT:
	  POP2(aw);
	  /*
	   * DOT is something of a kludge. If it's invoked while we're parsing
	   * a layout (I.E. a layout wants to reference some previous
	   * widget attribute) it goes about its work in a totally
	   * different way. This is necessary because of the way layouts
	   * are implemented at runtime (as special cased functions).
	   */
	  if (awl_current(aw)) {
	       /* We're in a layout */
	       if (value_type(a1) != STRING)
		    exec_error(aw, "Bad type '%s' for '.' LHS, not STRING",
			       type_name(value_type(a1)));
	       else if (value_type(a1) != STRING)
		    exec_error(aw, "Bad type '%s' for '.' RHS, not STRING",
			       type_name(value_type(a1)));
	       else {
		    Layout trav = wl_parent(awl_current(aw));

		    /* search up through the family tree for DOT LHS */
		    while (trav && strcmp(wl_name(trav), value_string(a1)))
			 trav = wl_parent(trav);
		    if (!trav)
			 exec_error(aw, "No layout named '%s' for '.' LHS",
				    value_string(a1));
		    else {
			 trav = wl_head(trav);
			 while (trav && strcmp(wl_name(trav),
					       value_string(a2)))
			      trav = wl_next(trav);
			 if (!trav)
			      exec_error(aw, "No member '%s' for '.' RHS",
					 value_string(a2));
			 else {
			      value_any(res) = (Generic)trav;
			      value_type(res) = LAYOUT;
			 }
		    }
	       }
	  }
	  else if (value_type(a1) != LAYOUT)
	       exec_error(aw, "Bad type '%s' for '.' LHS, not LAYOUT",
			  type_name(value_type(a1)));
	  else if (value_type(a2) != STRING)
	       exec_error(aw, "Bad type '%s' for '.' RHS, not STRING",
			  type_name(value_type(a2)));
	  else {
	       Value *v;

	       if (!(v = do_dot_lookup(value_any(a1), value_string(a2))))
		    exec_error(aw, "'%s' is not a member of '%s'",
			       value_string(a2), wl_name(value_any(a1)));
	       else
		    res = *v;
	  }
	  break;
  
     case EE:
          LOGICAL_BINARY_OP(aw, ==);
	  break;

     case GE:
	  LOGICAL_BINARY_OP(aw, >=);
	  break;

     case GT:
     	  LOGICAL_BINARY_OP(aw, >);
	  break;

     case LAND:
	  INTEGER_BINARY_OP(aw, &&);
	  break;

     case LE:
	  LOGICAL_BINARY_OP(aw, <=);
	  break;

     case LOR:
	  LOGICAL_BINARY_OP(aw, ||);
	  break;

     case LT:
	  LOGICAL_BINARY_OP(aw, <);
	  break;

     case MBR:
	  a2 = pop(aw);				/* symbol */
	  a1 = coerce(aw, pop(aw), ANY);	/* member */
	  value_clear(res);

	  if (value_type(a2) == SYMBOL && symbol_alist(value_symbol(a2))) {
	       sym = value_symbol(a2);
	       a0 = coerce(aw, a1, STRING);
	       if (symbol_alist_find(sym, a0, FALSE)) {
		    value_set(res, DATA, INT, int, TRUE);
		    value_type(a2) = 0; /* prevent further evaluation */
	       }
	       else
		    a2 = coerce(aw, a2, ANY);
	  }
	  else
	       a2 = coerce(aw, a2, ANY);
	  if (value_type(a2) == STRING) {
	       char *tmp = NULL;

	       if (value_type(a1) == CHAR) {
		    tmp = index(value_string(a2), value_char(a1));
		    if (tmp)
			 value_set(res, STATIC, STRING, aobj, new_aobj(tmp));
		    else
			 value_clear(res);
	       }
	       else if (value_type(a1) == STRING) {
		    tmp = s_strsed(value_string(a2), value_string(a1));
		    if (tmp)
			 value_set(res, DATA, STRING, aobj, new_aobj(tmp));
		    else
			 value_clear(res);
	       }
	       else
		    value_clear(res);
	  }
	  else if (value_type(a2) == RANGE) {
	       a1 = coerce(aw, a1, INT);
	       value_set(res, DATA, INT, int,
			 (value_int(a1) >= range_min(value_range(a2)) &&
			  value_int(a1) <= range_max(value_range(a2))));
	  }
	  else if (value_type(a2))
	       value_clear(res);
	  break;

     case MINUS:
	  INTEGER_UNARY_OP(aw, -);
	  break;

     case MOD:
	  POP2(aw);
	  value_type(res) = value_type(a1);
	  switch(value_type(a1)) {
	  case CHAR:
	       value_char(res) = value_char(a1)
		    % value_char(coerce(aw, a2, INT)); 
	       break;

	  case INT:
	       value_int(res) = value_int(a1)
		    % value_int(coerce(aw, a2, INT));
	       break;

	  case FLOAT:	/* % can't hack floats */
	       value_type(res) = INT;
	       value_int(res) = value_int(coerce(aw, a1, INT))
		    % value_int(coerce(aw, a2, INT));
	       break;

	  default:
	       value_type(res) = INT;
	       value_int(res) = 0;
	       break;
	  }
	  break;
	  
     case MULT:
          ARITH_OP(aw, *, mult_alloc);
	  break;

     case NE:
     	  LOGICAL_BINARY_OP(aw, !=);
     	  break;

     case NOT:
     	  INTEGER_UNARY_OP(aw, !);
     	  break;

     case NUKE:
	  a2 = pop(aw);
	  if (value_type(a2) != SYMBOL)
	       do_free(&a2);
	  else
	       symbol_remove(aw, value_symbol(a2));
	  break;

     case OR:
     	  INTEGER_BINARY_OP(aw, |);
     	  break;

     case PLUS:
	  ARITH_OP(aw, +, add_alloc);
	  break;

     case SL:
     	  INTEGER_BINARY_OP(aw, <<);
     	  break;

     case SR:
     	  INTEGER_BINARY_OP(aw, >>);
     	  break;

     case SUB:
          ARITH_OP(aw, -, sub_alloc);
	  break;

     case TRI:
	  a2 = pop(aw);				/* : part */
	  a1 = pop(aw);				/* ? part */
	  a0 = coerce(aw, pop(aw), ANY);	/* conditional */
	  if (is_allocated_type(value_type(a0)))
	       cond = (aobj_ptr(value_aobj(a0)) != NULL);
	  else
	       cond = (value_any(a0) != NULL);
	  if (cond)
	       res = a1;
	  else
	       res = a2;
	  break;

     case XOR:
     	  INTEGER_BINARY_OP(aw, ^);
	  break;

     default:
	  exec_error(aw, "Unknown ALU opcode");
	  break;
     }
     push(aw, res);
}


Local int asoc_value(aw, val, a1, valp)
AwlWidget aw;
Value val, a1;
Value *valp;
{
     int typ = value_type(val);
     int idx = value_type(a1) == CHAR ? value_char(a1) :
	  value_type(a1) == INT ? value_int(a1) : -1;
	  

     /*
      * If the it's a string or list and the value is a valid
      * INT/CHAR subscript for it, assume that it's an index op.
      */
     if (typ == STRING && idx >= 0)
	  value_set(*valp, DATA, CADDR, any, value_string(val) + idx);
     else if (typ == LIST && idx >= 0)
	  value_set(*valp, DATA, SADDR, any, value_list(val) + idx);
     else if (typ == LAYOUT || typ == WIDGET) {
	  Generic tmp = (typ == LAYOUT ? value_any(val) : NULL);

	  if (value_type(a1) != STRING) {
	       exec_warn(aw, "Resource name index not a string");
	       return 0;
	  }
	  else
	       value_set(*valp, DATA, RESOURCE, any,
			 res_link(aw, tmp, value_string(a1), val));
     }
     else {
	  value_clear(*valp);
	  return 0;
     }
     return 1;
}

/* Widget layout manipulation functions */

/* open a new declaration level */
Local void open_w(aw)
AwlWidget aw;
{
     Value a1, a2;

     POP2(aw);
     if (value_type(a1) != WCLASS)
	  exec_error(aw, "Widget type not WidgetClass");
     else if (value_type(a2) != STRING)
	  exec_error(aw, "Widget name not a string");
     else {
	  if (!awl_current(aw))
	       awl_current(aw) = new_wl(value_wclass(a1));
	  else {
	       Layout p1;

	       p1 = new_wl(value_wclass(a1));
	       if (!wl_head(awl_current(aw)))
		    wl_head(awl_current(aw)) = wl_tail(awl_current(aw)) = p1;
	       else {
		    wl_next(wl_tail(awl_current(aw))) = p1;
		    wl_tail(awl_current(aw)) = p1;
	       }
	       wl_parent(p1) = awl_current(aw);
	       awl_current(aw) = p1;
	  }
	  wl_name(awl_current(aw)) = value_string(a2);
     }
}

Local Resource res_link(aw, l, name, val)
AwlWidget aw;
Layout l;
String name;
Value val;
{
     Resource tmp = NULL;

     if (l)
	  tmp = wl_resources(l);
     while (tmp) {
	  if (!strcomp(res_name(tmp), name)) {
	       res_value(tmp) = val;
	       break;
	  }
	  else
	       tmp = res_next(tmp);
     }
     if (!tmp) {
	  tmp = new_res(l);
	  res_name(tmp) = name;
	  res_value(tmp) = val;
	  if (l) {
	       res_next(tmp) = wl_resources(l);
	       wl_resources(l) = tmp;
	  }
     }
     return(tmp);
}

Local void ares_w(aw)
AwlWidget aw;
{
     Value a1, a2;

     if (!awl_current(aw))
	  exec_error(aw, "Resource not in widget declaration");
     else {
	  POP2(aw);
	  if (value_type(a2) != STRING)
	       exec_error(aw, "resource type not STRING");
	  res_link(aw, awl_current(aw), value_string(a2), a1);
     }
}

Local void close_w(aw)
AwlWidget aw;
{
     if (!awl_current(aw))
	  exec_error(aw, "Mismatched CLOS encountered");
     else {
	  awl_active(aw) = awl_current(aw);
	  awl_current(aw) = wl_parent(awl_current(aw));
     }
}

/*
 * These two routines deal with possible floating point/math errors. Because
 * we can't pass arguments to signal routines, the global widget value
 * "Executor" is set by alu() for use when and if a signal occurs.
 */

#if defined(SIGFPE)
/* Deal with floating point exceptions */
Local void fp_exception()
{
#if defined(SYSV) || defined(USG) || defined(usg)
     signal(SIGFPE, fp_exception);
#endif
     exec_warn(Executor, "floating point exception");
}

#endif /* SIGFPE */

#if defined(SIGZERO)
/* Deal with zero divides */
Local void zero_divide()
{
#if defined(SYSV) || defined(USG) || defined(usg)
     signal(SIGZERO, zero_divide);
#endif
     exec_warn(Executor, "zero divide");
}
#endif /* SIGZERO */
