#ifndef lint
static char *rcsid = "$Header: /usr/src/local/awl/RCS/evalutils.c,v 2.5 90/04/19 20:05:07 jkh Exp $";
#endif

/*
 *
 *                   Copyright 1989, 1990
 *                    Jordan K. Hubbard
 *
 *		  PCS Computer Systeme, GmbH.
 *	             Munich, West Germany
 *
 *
 * This file is part of AWL.
 *
 * AWL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * AWL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AWL; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

/*
 * This file contains various esoteric utility routines for the evaluator.
 *
 * $Log:	evalutils.c,v $
 * Revision 2.5  90/04/19  20:05:07  jkh
 * Alpha checkin.
 * 
 * Revision 2.4  90/03/31  05:41:07  jkh
 * Fixed stupid typo in op_nargs() that broke the world.
 * 
 * Revision 2.3  90/03/30  21:59:43  jkh
 * Changed the way opcodes are searched for.
 * 
 * Revision 2.2  90/03/29  18:28:42  jkh
 * Rearranged some code.
 * 
 * Revision 2.1  90/03/27  09:00:07  jkh
 * Moved do_convert() to convert.c
 * 
 * Revision 2.0  90/03/26  01:43:12  jkh
 * pre-beta check-in
 *
 */

#include "AwlP.h"
#include "y.tab.h"

/* Kludge needed by yyparse() during parsing */
Export AwlWidget _currW;

/* Kludge needed by signal handler */
Export Symbol *_procSignalTable;

/*
 * Return the number of args for an opcode. We don't have to bsearch for
 * the value since the opcodes are grouped cleverly together to maximize
 * this one case (look at op_name() to see the default method).
 */
Export Inline int op_nargs(insn)
int insn;
{
     if (insn < ARES)
	  return 1;
     if (insn < PUSH)
	  return 0;
     return 2;
}

/* Do the munging and frobbing necessary to do <expr>.<expr> lookups */
Export Value *do_dot_lookup(s1, n)
Layout s1;
String n;
{
     Local Value val;

     if (!wl_head(s1))
	  return(NULL);
     else {
	  s1 = wl_head(s1);
	  while (s1) {
	       if (!strcomp(n, wl_name(s1))) {
		    value_set(val, DATA, LAYOUT, any, (Generic)s1);
		    return(&val);
	       }
	       s1 = wl_next(s1);
	  }
	  return(NULL);
     }
}

/* Generate unique labels for compiler generated branches */
Export Symbol do_new_label(aw, addr)
AwlWidget aw;
int addr;
{
     static int lno = 0;
     char name[10];
     Symbol tmp;

     sprintf(name, "__L%d", lno++);
     tmp = symbol_new(aw, name, INT);
     symbol_class(tmp) = LABEL;
     symbol_type(tmp) = INT;
     symbol_int(tmp) = addr;
     symbol_add(tmp);
     return(tmp);
}

/*
 * Adjust a segment by either growing or shrinking the data storage allotted
 * to it. This does not change the segment position pointer, it should
 * be adjusted by the caller if so desired.
 */
Export Inline void adjust(aw, ptr, amt)
AwlWidget aw;
Generic ptr;
int amt;
{
     if (ptr == (Generic)awl_stack(aw)) {
	  if (amt > 0 && (awl_sp(aw) + amt) >= awl_estack(aw)) {
	       awl_estack(aw) += (STACK_INCR <= amt ? STACK_INCR : amt);
	       awl_stack(aw) = (Value *)XtRealloc(awl_stack(aw),
						  awl_estack(aw) *
						  sizeof(Value));
	  }
	  else if (amt < 0)
	       exec_error(aw, "Attempt to shrink stack segment");
	  /*
	   * See if we can reclaim a little space. We do this when the
	   * difference between the stack pointer and the end of stack
	   * is greater than 2 "chunks" worth of stack. We only want
	   * about one chunk worth of float.
	   */
	  if (awl_estack(aw) - awl_sp(aw) > (STACK_INCR * 2)) {
	       awl_estack(aw) -= STACK_INCR;
	       awl_stack(aw) = (Value *)XtRealloc(awl_stack(aw),
						  awl_estack(aw) *
						  sizeof(Value));
	  }
     }
     else if (ptr == (Generic)awl_text(aw)) {
	  if (amt > 0 && (awl_lastpc(aw) + amt) >= awl_etext(aw)) {
	       awl_etext(aw) += (TEXT_INCR <= amt ? TEXT_INCR : amt); 
	       awl_text(aw) = (Datum *)XtRealloc(awl_text(aw),
						 awl_etext(aw) *
						 sizeof(Datum));
	  }
	  else if (amt < 0)
	       exec_error(aw, "Attempt to shrink text segment");
     }
     else if (ptr == (Generic)awl_ltab(aw)) {
	  if (amt > 0 && (awl_ltabcurr(aw) + amt) >= awl_ltabsize(aw)) {
	       awl_ltabsize(aw) += (LTAB_INCR <= amt ? LTAB_INCR : amt); 
	       awl_ltab(aw) = (LTabEnt *)XtRealloc(awl_ltab(aw),
						   awl_ltabsize(aw) *
						   ltab_size);
	  }
	  else if (amt < 0)
	       exec_error(aw, "Attempt to shrink line table");
     }
     else
	  exec_error(aw, "Non-existant segment referenced in adjust");
}

/* free up storage for value if necessary */
Export Inline void do_free(val)
Value *val;
{
     if (is_allocated_type(value_type(*val))
	 && !is_precious(value_class(*val))) {
	  --aobj_cnt(value_aobj(*val));
	  if (value_type(*val) == STRING) {
	       if (aobj_cnt(value_aobj(*val)) <= 0) {
		    if (value_string(*val) && value_class(*val) != STATIC)
			 XtFree(value_string(*val));
		    XtFree(value_aobj(*val));
		    value_clear(*val);
	       }
	  }
	  else if (value_type(*val) == LIST) {
	       if (aobj_cnt(value_aobj(*val)) <= 0) {
		    if (value_list(*val) && value_class(*val) != STATIC)
			 list_free(value_list(*val));
		    XtFree(value_aobj(*val));
		    value_clear(*val);
	       }
	  }
     }
}

Local void exec_msg(aw, type, s, p1, p2, p3)
AwlWidget aw;
int type;
String s;
Generic p1; /* don't move these to one line or the define */
Generic p2; /* will ream you good */
Generic p3;
{
     Local msg[128], msg_line[512];
     LTabEnt *lt;

     if ((lt = ltab_for_pc(aw, awl_pc(aw))) != NULL)
	  sprintf(msg, "%s:%d: %s", ltab_file(*lt), ltab_line(*lt), s);
     else
	  sprintf(msg, "At $PC %#x: %s", awl_pc(aw), s);
     sprintf(msg_line, msg, p1, p2, p3);
     fflush(stdout); fflush(stderr);
     if (type == -1)
	  fprintf(awl_err(aw), "Debug: %s\n", msg_line);
     else if (type == 0)
	  XtWarning(msg_line);
     else
	  XtError(msg_line);
}

Export Inline void exec_error(aw, s, p1, p2, p3)
AwlWidget aw;
String s;
Generic p1;  /* here too */
Generic p2; 
Generic p3;
{
     Import void longjmp();

     exec_msg(aw, 1, s, p1, p2, p3);
     longjmp(awl_bomb(aw), 1);
}

Export Inline void exec_warn(aw, s, p1, p2, p3)
AwlWidget aw;
String s;
Generic p1; /* same here */
Generic p2;
Generic p3;
{
     if (!awl_nowarn(aw))
	  exec_msg(aw, 0, s, p1, p2, p3);
}

Export Inline void debug(aw, s, p1, p2, p3)
AwlWidget aw;
String s;
Generic p1; /* nope, not here either */
Generic p2;
Generic p3;
{
     if (awl_debug(aw))
	  exec_msg(aw, -1, s, p1, p2, p3);
}

/* Front-end a signal for a user defined (or awl) function */
Export void do_signal_user(sig)
int sig;
{
     Value val;
     Symbol sym;

     if (!_procSignalTable || !(sym = _procSignalTable[sig])) {
	  fprintf(stderr, "No signal handler for signal %d!\n", sig);
	  exit(1);
     }
     /* Ignore further signals until we deal with this one properly */
     signal(sig, SIG_IGN);
     do_link(symbol_root(sym));
     value_set(val, DATA, INT, int, sig);
     push(symbol_root(sym), val);
     Eval(sym);
     /* Ok, now we can put the signal handler back */
     signal(sig, do_signal_user);
}

/* Front-end a callback request for a user defined (or awl) function */
Export void do_callback_user(w, client_data, call_data)
Widget w;
Generic client_data;
Generic call_data;
{
     Value val;
     Symbol tmp;
     AwlWidget aw;

     tmp = (Symbol)client_data;
     aw = symbol_root(tmp);
     do_link(aw);
     value_set(val, DATA, WIDGET, widget, w);
     push(aw, val);
     value_set(val, DATA, SYMBOL, symbol, client_data);
     push(aw, val);
     value_set(val, DATA, ANY, any, call_data);
     push(aw, val);
     Eval(tmp);
}

/*
 * Front-end a timer callback for a user defined function.
 */
Export void do_timer_user(client_data, id)
Generic client_data;
XtIntervalId *id;
{
     Value val;
     Symbol tmp;
     AwlWidget aw;

     tmp = (Symbol)client_data;
     aw = symbol_root(tmp);
     do_link(aw);
     value_set(val, DATA, SYMBOL, symbol, tmp);
     push(aw, val);
     value_set(val, DATA, ANY, any, id ? (Generic)*id : NULL);
     push(aw, val);
     Eval(tmp);
}

/*
 * Front-end a work procedure callback for a user defined function.
 */
Export void do_workproc_user(client_data)
XtPointer client_data;
{
     Symbol tmp;

     tmp = (Symbol)client_data;
     do_link(symbol_root(tmp));
     Eval(tmp);
}

/*
 * Given a widget, find the AWL top.
 */
Local AwlWidget find_awl_top(w)
Widget w;
{
     Widget tmp = w;

     while (tmp && XtClass(tmp) != awlWidgetClass)
	  tmp = XtParent(tmp);
     if (!tmp) {
	  fprintf(stderr, "find_awl_top: Widget %x not child of awl widget!\n",
		  w);
	  exit(1);
     }
     return((AwlWidget)tmp);
}

/*
 * Front-end a translation callback request for a user defined function.
 */
Export void do_action_user(w, event, argv, argc)
Widget w;
XEvent *event;
String *argv;
int *argc;
{
     Symbol tmp;
     Value val;
     AwlWidget aw = find_awl_top(w);

     if (*argc < 1)
	  exec_error(aw, "do_action_user: No function name!");
     else if ((tmp = symbol_find(aw, argv[0])) == NULL)
	  exec_error(aw, "do_action_user: No such function '%s'", argv[0]);
     else {
	  int i;

	  do_link(aw);
	  value_set(val, DATA, WIDGET, widget, w);
	  push(aw, val);
	  for (i = 1; i < *argc; i++) {
	       value_set(val, STATIC, STRING, aobj, new_aobj(argv[i]));
	       push(aw, val);
	  }
	  Eval(tmp);
     }
}

/*
 * Get an argument of a given "type" from the stack. If "check" is
 * true, complain if stack argument isn't of "type", otherwise
 * convert as necessary.
 */
Export Inline Value get_arg(aw, argnum, type, check)
AwlWidget aw;
int argnum, type;
Boolean check;
{
     Value val;

     if (argnum > awl_nparms(aw))
	  exec_warn(aw, "get_arg: arg #%d > num args (%d)", argnum,
		     awl_nparms(aw));
     
     val = awl_stack(aw)[FRAME_OFFSET(-argnum)];

     if (check) {
	  if (value_type(val) == type)
	       return(val);
	  else
	       val = do_load(aw, val, FALSE);
	  if (value_type(val) != type) {
	       exec_warn(aw, "Type mismatch: Expected '%s', got '%s'",
			  type_name(type), type_name(value_type(val)));
	       value_type(val) = 0;
	  }
     }
     else
	  val = coerce(aw, val, type);
     return(val);
}

/* ancillary string and list functions */
Export AllocObj add_alloc(aw, a1, a2)
AwlWidget aw;
Value a1, a2;
{
     Value tmp;
     String ret, *lst;
     int i, j, l1, l2;

     if (value_type(a1) == STRING) {
	  tmp = coerce(aw, a2, STRING);
	  ret = (Generic)XtMalloc(strlen(value_string(a1))
				  + strlen(value_string(tmp)) + 1);
	  strcpy(ret, value_string(a1));
	  strcat(ret, value_string(tmp));
     }
     else if (value_type(a1) == LIST) {
	  tmp = coerce(aw, a2, LIST);
	  l1 = list_len(value_list(a1));
	  l2 = list_len(value_list(tmp));
	  lst = (String *)XtMalloc((l1 + l2 + 1) * sizeof(String *));
	  j = 0;
	  for (i = 0; i < l1; i++)
	       lst[j++] = XtNewString(value_list(a1)[i]);
	  for (i = 0; i < l2; i++)
	       lst[j++] = XtNewString(value_list(tmp)[i]);
	  lst[j] = NULL;
	  ret = (String)lst;
     }
     return(new_aobj(ret));
}

Export AllocObj sub_alloc(aw, a1, a2)
AwlWidget aw;
Value a1, a2;
{
     AllocObj ret = NULL;
     String tmp0, tmp1, tmp2;
     int orig_type;

     /* first, make sure we're subtracting from a string */
     tmp0 = (value_type(a1) == LIST) ? list_to_string(aw, value_list(a1))
	  : value_string(a1);
     /*
      * now make sure what we're subtracting is also a string but
      * first save its original type in case it was also a list (in which
      * case we don't want to add $SEP since it's already been done in the
      * conversion).
      */
     orig_type = value_type(a2);
     a2 = coerce(aw, a2, STRING);
     /* construct a temporary strsed "command" */
     tmp1 = XtMalloc(strlen(value_string(a2)) + strlen(awl_sep(aw)) + 4);
     tmp1[0] = '/'; tmp1[1] = '\0';
     strcat(tmp1, value_string(a2));
     if (value_type(a1) == LIST && orig_type != LIST)
	  strcat(tmp1, awl_sep(aw));
     strcat(tmp1, "//");
     /* do it and free the temporary command */
     tmp2 = strsed(tmp0, tmp1, 0);
     XtFree(tmp1);
     /*
      * If the original "string" was actually a list, convert it back
      * and clean up after ourselves.
      */
     if (value_type(a1) == LIST) {
	  if (tmp2) {
	       ret = new_aobj(list_from_string(aw, tmp2));
	       XtFree(tmp2);
	  }
	  XtFree(tmp0);
     }
     else if (tmp2)
	  ret = new_aobj(tmp2);
     return(ret);
}

Export AllocObj mult_alloc(aw, a1, a2)
AwlWidget aw;
Value a1, a2;
{
     AllocObj ret = NULL;
     int m;
     String tmp0, tmp1;

     m = value_int(coerce(aw, a2, INT));
     if (m) {
	  tmp0 = (value_type(a1) == LIST) ? list_to_string(aw, value_list(a1))
	       : value_string(a1);
	  tmp1 = string_multiply(tmp0, m);
	  if (value_type(a1) == LIST) {
	       ret = new_aobj(list_from_string(aw, tmp1));
	       XtFree(tmp0);
	       XtFree(tmp1);
	  }
	  else
	       ret = new_aobj(tmp1);
     }
     return(ret);
}

/* NO-OP */
Export AllocObj div_alloc(aw, a1, a2)
AwlWidget aw;
Value a1, a2;
{
     return(value_aobj(a1));
}

/* These two routines handle loading of awl code */

Local String setup_cpp_args(aw)
AwlWidget aw;
{
     Import RMap ResMap[];
     register int i, rsize;
     String tmp, incdir;

#if defined(MOTIF)
     Local String wset = "-DMOTIF ";
#elif ATHENA
     Local String wset = "-DATHENA ";
#elif XW
     Local String wset = "-DXW ";
#else
     Local String wset = "";
#endif

     incdir = XtMalloc(strlen(AWLDIR) + 3);
     strcpy(incdir, "-I");
     strcat(incdir, AWLDIR);
     rsize = strlen(wset) + strlen(incdir) + 1;
     for (i = 0; ResMap[i].from; i++)
	  rsize += (strlen(ResMap[i].from) + strlen(ResMap[i].to) + 8);
     tmp = XtMalloc(rsize + 1);
     strcpy(tmp, wset);
     strcat(tmp, incdir);
     for (i = 0; ResMap[i].from; i++)
	  sprintf(tmp + strlen(tmp), " -D%s='\"%s\"'",
		  ResMap[i].from, ResMap[i].to);
     XtFree(incdir);
     return(tmp);
}

Export int parseFile(aw, fname)
AwlWidget aw;
String fname;
{
     int return_status = 0, flen = 0;
     Import FILE *yyin;
     String cmd;
     String args;

     args = setup_cpp_args(aw);
     if (fname)
	  flen = strlen(fname);
     cmd = XtMalloc(strlen(args) + strlen(awl_cpp_cmd(aw)) + flen + 40);
     if (flen)
	  sprintf(cmd, "%s %s %s",
		  awl_cpp_cmd(aw),
		  args,
		  fname);
     else
	  sprintf(cmd, "%s %s",
		  awl_cpp_cmd(aw),
		  args);

     /* cpp away.. */
     if (awl_debug(aw)) {
	  char tee[80];

	  strcpy(tee, "|tee ");
	  strcat(tee, tmpnam(NULL));
	  strcat(cmd, tee);
     }
     strcat(cmd, "\n");
     debug(aw, "Executing 'cpp' command : %s", cmd);
     if ((yyin = popen(cmd, "r")) == NULL) {
          exec_error(aw, "awl: parseFile: Can't invoke: %s", cmd);
	  return_status = 1;
     }
     else {
	  _currW = aw;
	  lexer_reset();
	  return_status = yyparse();
	  pclose(yyin);
	  /* clean up the parsers act */
	  if (yy_file) {
	       XtFree(yy_file);
	       yy_file = 0;
	  }
	  yy_line = 0;
     }
     XtFree(args);
     XtFree(cmd);
     return(return_status);
}
