#ifndef lint
static char *rcsid = "$Header: /usr/src/local/awl/RCS/layout.c,v 2.1 90/04/19 20:05:12 jkh Exp $";
#endif

/*
 *
 *                   Copyright 1989, 1990
 *                    Jordan K. Hubbard
 *
 *		  PCS Computer Systeme, GmbH.
 *	             Munich, West Germany
 *
 *
 * This file is part of AWL.
 *
 * AWL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * AWL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AWL; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

/*
 * This file contains the code for creating and inspecting widget
 * layouts and interfacing with the X toolkit's conversion routines.
 *
 * $Log:	layout.c,v $
 * Revision 2.1  90/04/19  20:05:12  jkh
 * Alpha checkin.
 * 
 * Revision 2.0  90/03/26  01:44:33  jkh
 * pre-beta check-in
 * 
 */

#include "AwlP.h"
#include "y.tab.h"

/*
 * Given a value, figure out how to set a given "arg" element
 * to it.
 */
Export void wset_arg_by_value(args, idx, name, val)
Arg *args;
int *idx;
String name;
Value val;
{
     switch (value_type(val)) {
     case FLOAT:
	  /* floats are passed by address for portability */
	  XtSetArg(args[*idx], name, (XtArgVal)&value_float(val));
	  ++(*idx);
	  break;

	  /* usually a widget ID that was pre-ref'd */
     case WIDGET:
	  if (value_class(val) == INDIRECT) {
	       XtSetArg(args[*idx], name, *(Widget *)value_widget(val));
	       ++(*idx);
	  }
	  else if (value_class(val) == DATA) {
	       XtSetArg(args[*idx], name, (XtArgVal)value_widget(val));
	       ++(*idx);
	  }
	  else
	       exec_error("Illegal class for resource typed WIDGET");
	  break;

     case CHAR:
	  XtSetArg(args[*idx], name, (XtArgVal)value_char(val));
	  ++(*idx);
	  break;

     AOBJCASE:
	  XtSetArg(args[*idx], name, (XtArgVal)aobj_ptr(value_aobj(val)));
	  ++(*idx);
	  break;
	  
	  /* All else can be passed in "any" */
     default:
	  XtSetArg(args[*idx], name, (XtArgVal)value_any(val));
	  ++(*idx);
	  break;
     }
}

/*
 * Given a widget class and a "manage" flag, decide what toolkit function
 * should be called to create a given widget.
 */
Export Widget (*wcreate_rtn(class, manage))()
WidgetClass class;
Boolean manage;
{
     int flags = map_wclass_to_flags(class);

     if (flags & W_WIDPROC)
	  return((Widget(*)())class);
     else if (flags & W_POPUP)
	  return(XtCreatePopupShell);
     else if (manage)
	  return(XtCreateManagedWidget);
     else
	  return(XtCreateWidget);
}

/* return the Widget ID for a layout, triggering an error if none */
Export Inline Widget layout_id(aw, n)
AwlWidget aw;
Layout n;
{
     if (!n)
	  exec_error(aw, "layout_id: Null layout passed");
     else if (!wl_id(n))
	  exec_error(aw, "layout_id: Layout %s has not yet been created",
		     wl_name(n));
     else
	  return(wl_id(n));
     /* Not reached, but keeps gcc happy */
     return((Widget)0);
}

/* return the Widget class for a layout, triggering an error if none */
Export Inline WidgetClass layout_wclass(aw, n)
AwlWidget aw;
Layout n;
{
     if (!n)
	  exec_error(aw, "layout_wclass: Null layout passed");
     else if (!wl_wclass(n))
	  exec_error(aw, "layout_wclass: Layout %s has not yet been created",
		     wl_name(n));
     else
	  return(wl_wclass(n));
     /* Not reached, but keeps gcc happy */
     return((WidgetClass)0);
}

/*
 * Create a widget for a layout.
 */
Export void create_widget(aw, parent, n, manage)
AwlWidget aw;
Widget parent;
Layout n;
Boolean manage;
{
     int a, flags;
     Arg args[50];
     Resource r;
     Generic map;

     a = 0;
     r = wl_resources(n);

     while (r) {
	  if ((map = map_rname(res_name(r), &flags)) != NULL) {
	       if (flags & W_RESPROC)
		    (*(void (*)())map)(aw, args, &a, r);
	       else
		    wset_arg_by_value(args, &a, (String)map, res_value(r));
	  }
	  else
	       wset_arg_by_value(args, &a, res_name(r), res_value(r));
	  r = res_next(r);
     }

     wl_id(n) = ((*wcreate_rtn)(wl_wclass(n), manage))(wl_name(n),
						       wl_wclass(n),
						       parent, args, a);
     if (!wl_id(n)) {
	  exec_error(aw, "Creating widget '%s'", wl_name(n));
	  return;
     }
}

Export void layout_create(aw, parent, w, manage)
AwlWidget aw;
Widget parent;
Layout w;
Boolean manage;
{
     while (w) {
	  create_widget(aw, parent, w, manage);
	  if (wl_head(w)) {
	       if (map_wclass_to_flags(wl_wclass(w)) & W_MANAGER)
		    layout_create(aw, wl_id(w), wl_head(w), manage);
	       else
		    exec_warn(aw,
			      "%s: Widget '%s' Not MANAGER: Ignoring children",
			      "xtcreate", wl_name(w),
			      map_wclass_to_wname(wl_wclass(w)));
	  }
	  w = wl_next(w);
     }
}
