#ifndef _LTYPES_H_INCLUDE
#define _LTYPES_H_INCLUDE

/* $Header: /usr/src/local/awl/RCS/ltypes.h,v 2.2 90/04/19 20:05:20 jkh Exp $ */

/*
 *
 *                     Copyright 1989
 *                    Jordan K. Hubbard
 *
 *		  PCS Computer Systeme, GmbH.
 *	             Munich, West Germany
 *
 *
 * This file is part of AWL.
 *
 * AWL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * AWL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AWL; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

/*
 * Here are all the awl internal / external types.
 *
 * $Log:	ltypes.h,v $
 * Revision 2.2  90/04/19  20:05:20  jkh
 * Alpha checkin.
 * 
 * Revision 2.1  90/03/31  06:38:28  jkh
 * Eliminated some dead typedefs.
 * 
 * Revision 2.0  90/03/26  01:43:19  jkh
 * pre-beta check-in
 *
 */

typedef hash_table *Table;	/* as defined by tsearch and friends */

typedef struct _insntab {	/* instruction table */
     int insn;			/* insn # */
     int nargs;			/* number of operands */
     String name;		/* print name */
} InsnTable;

typedef struct _ltabent {
     String file;		/* file name */
     int line;			/* line number */
     int start;			/* PC of first insn for line */
     int last;			/* PC of last insn for line */
} LTabEnt;

typedef struct _breakpoint {
     struct _breakpoint *next;	/* next breakpoint in chain */
     int (*func)();		/* function to call on break */
     int addr;			/* address of breakpoint */
     Boolean enable;		/* enable state */
} *Breakpoint;

typedef struct _ctable {
     unsigned long *colors;	/* colors array */
     Table htab;		/* color name hash table */
     int nch;			/* # of chars per pixel */
     int ncolors;		/* number of colors */
} *CTable;

typedef struct _rmap {
     String from;		/* map from -> to */
     String to;
} RMap;

typedef struct _process {
     int pid;			/* process ID */
     FILE *in, *out;		/* input/output fd's */
} *Process;

typedef struct _allocobj {
     int cnt;			/* ref count */
     Generic ptr;		/* data */
} *AllocObj;

typedef struct _sym {
     AwlWidget aw;		/* which widget we belong to */
     String name;		/* symbol name */
     Table alist;		/* symbol's a-list (if any) */
     Table parent;		/* symbol's parent table */
     struct _sym *next;		/* for relating groups of symbols */
     struct _val {		/* symbol's value */
	  short class;		/* value's class */
	  short type;		/* value's type */
	  union _datum {	/* data */
	       AllocObj o;	/* data is allocated object */
	       FILE *z;		/* data is FILE */
	       FTYPE f;		/* data is float */
	       Widget w;	/* data is widget */
	       WidgetClass c;	/* data is widget class */
	       Generic a;	/* data is any (generic) */
	       char b;		/* data is char */
	       int i;		/* data is int */
	       short r[2];	/* data is range */
	       struct _sym *y;	/* data is symbol pointer */
	  } v;
     } v;
} *Symbol;

typedef struct _val Value;
typedef union _datum Datum;

typedef Generic (*GenFunc)();

typedef struct _function {
     int nparms;		/* number of function parameters */
     int nlocals;		/* number of function locals */
     int end;			/* ending PC address */
     Generic calldata;		/* hook for function specific data */
     union _addr {
	  int u;		/* "text" addr of user function */
	  Value (*s)();		/* function address of built-in function */
	  GenFunc f;		/* function address of external function */
     } addr;
} *Function;

typedef union _addr Address;

typedef struct _layout {
     Widget w_id;		/* the Widget ID */
     String name;		/* widget's name */
     WidgetClass class;		/* class */
     struct _layout *parent;	/* daddy */
     struct _layout *head, *tail;/* children */
     struct _layout *next;	/* siblings */
     struct _resource {
	  String name;
	  Value val;
	  struct _layout *parent;
	  struct _resource *next;
     } *res;
} *Layout;

typedef struct _resource *Resource;

#endif /* _LTYPES_H_INCLUDE */
