#ifndef lint
static char *rcsid = "$Header: /usr/src/local/awl/RCS/mathrtns.c,v 2.1 90/04/19 20:05:24 jkh Exp $";
#endif

/*
 *
 *                     Copyright 1989
 *                    Jordan K. Hubbard
 *
 *		  PCS Computer Systeme, GmbH.
 *	             Munich, West Germany
 *
 *
 * This file is part of AWL.
 *
 * AWL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * AWL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AWL; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

/*
 * This file contains all the math specific built-ins and system dependent
 * error handlers for under/overflow traps.
 *
 * $Log:	mathrtns.c,v $
 * Revision 2.1  90/04/19  20:05:24  jkh
 * Alpha checkin.
 * 
 * Revision 2.0  90/03/26  01:43:29  jkh
 * pre-beta check-in
 * 
 */

#include "AwlP.h"
#include "y.tab.h"

Local Value arg;

#define DEFUN_SIMPLE_MATH_FUNC(func) \
DEFUN(func) \
{ \
     arg = get_arg(aw, 1, FLOAT, FALSE); \
     value_float(arg) = func(value_float(arg)); \
     return(arg); \
}

#if defined(__STDC__) && !defined(UNIXCPP)	/* ANSI */
#define DEFUN_NONEG_MATH_FUNC(func) \
DEFUN(func) \
{ \
     arg = get_arg(aw, 1, FLOAT, FALSE); \
     if (value_float(arg) < 0) \
	  exec_warn(aw, "Attempt to compute %s of negative number.", #func); \
     else \
	  value_float(arg) = sqrt(value_int(arg)); \
     return(arg); \
}
#else
DEFUN(func) \
{ \
     arg = get_arg(aw, 1, FLOAT, FALSE); \
     if (value_float(arg) < 0) \
	  exec_warn(aw, "Attempt to compute %s of negative number.", "func"); \
     else \
	  value_float(arg) = sqrt(value_int(arg)); \
     return(arg); \
}
#endif

/*
 * sqrt(X):
 * This gives you the positive square root of X.  It reports an
 * error if X is negative.
 */
DEFUN_NONEG_MATH_FUNC(sqrt)

/*
 * log(X):
 * This gives you the natural logarithm of X. It reports an error
 * if X is negative.
 */
DEFUN_NONEG_MATH_FUNC(log)

/*
 * log10(X):
 * This gives you the logarithm base 10 of X. It reports and error
 * if X is negative.
 */
DEFUN_NONEG_MATH_FUNC(log10)

/*
 * exp(X):
 * This gives you the exponential of X, or reports an error if X is
 * out of range.  The range of values X can have depends on your
 * machine's floating point representation.
 */
DEFUN_SIMPLE_MATH_FUNC(exp)

/*
 * sin(X):
 * This gives you the sine of X, with X in radians.
 */
DEFUN_SIMPLE_MATH_FUNC(sin)

/*
 * cos(X):
 * This gives you the cosine of X, with X in radians.
 */
DEFUN_SIMPLE_MATH_FUNC(cos)

/*
 * tan(x):
 * This gives you the tangent of X, with X in radians.
 */
DEFUN_SIMPLE_MATH_FUNC(tan)

/*
 * asin(x):
 * This gives you the arcsine of X, with X in radians.
 */
DEFUN_SIMPLE_MATH_FUNC(asin)

/*
 * acos(x):
 * This gives you the arccosine of X, with X in radians.
 */
DEFUN_SIMPLE_MATH_FUNC(acos)

/*
 * atan(x):
 * This gives you the arctangent of X, with X in radians.
 */
DEFUN_SIMPLE_MATH_FUNC(atan)

/*
 * atan2(Y, X):
 * This gives you the arctangent of Y/X, with both in radians.
 */
DEFUN(atan2)
{
     Value y, x;

     y = get_arg(aw, 1, FLOAT, FALSE);
     x = get_arg(aw, 2, FLOAT, FALSE);
     value_float(arg) = atan2(value_float(y), value_float(x));
     return(arg);
}

/*
 * hypot(X, Y):
 * Inplements the standard Euclidean distance function.
 */
DEFUN(hypot)
{
     Value x, y;

     x = get_arg(aw, 1, FLOAT, FALSE);
     y = get_arg(aw, 2, FLOAT, FALSE);
     value_float(arg) = hypot(value_float(x), value_float(y));
     return(arg);
}

/*
 * pow(X, Y):
 * Pow returns XY.  If X is zero, Y must be positive.  If X is
 * negative, Y must be an integer.
 */
DEFUN(pow)
{
     Value x, y;

     x = get_arg(aw, 1, FLOAT, FALSE);
     y = get_arg(aw, 2, ANY, FALSE);
     value_set(arg, DATA, FLOAT, float, 0.0);
     if (value_type(y) == FLOAT) {
	  if (value_float(x) == 0.0 && value_float(y) < 0.0)
	       exec_warn(aw, "pow: If X is 0.0, Y must be >= 0.0");
	  else if (value_float(x) < 0.0)
	       exec_warn(aw, "pow: If X is less than 0.0, Y must be INT");
	  else
	       value_set(arg, DATA, FLOAT, float, pow(value_float(x),
						      value_float(y)));
     }
     else {
	  y = coerce(aw, y, INT);
	  if (value_float(x) == 0.0 && value_int(y) < 0)
	       exec_warn(aw, "pow: If X is 0.0, Y must be >= 0");
	  else
	       value_set(arg, DATA, FLOAT, float, pow(value_float(x),
						      (float)value_int(y)));
     }
     return(arg);
}
	  
/*
 * rand():
 * This gives you a random number between 0 and 1.
 */
DEFUN(rand)
{
     Import float drand48();

     value_type(arg) = FLOAT;
     value_float(arg) = drand48();
     return(arg);
}

/*
 * srand(X):
 * Sets the random number generator seed to X.
 */
DEFUN(srand)
{
     Import void srand48();

     arg = get_arg(aw, 1, INT, FALSE);
     srand48(value_int(arg));
     return(arg);
}
