#ifndef lint
static char *rcsid = "$Header: /usr1/src/local/awl/RCS/stat.c,v 2.1 90/04/19 20:05:44 jkh Exp Locker: jkh $";
#endif lint

/*
 *
 *                     Copyright 1990
 *                    Jordan K. Hubbard
 *
 *		  PCS Computer Systeme, GmbH.
 *	             Munich, West Germany
 *
 *
 * This file is part of AWL.
 *
 * AWL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * AWL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AWL; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

/*
 * Return verbose description of stat struct. I'm not completely sure
 * that this Does The Right Thing for all cases under BSD and SYSV, but
 * I've done my best to handle the differences I know about. If you find
 * any holes, please tell me.
 *
 * $Log:	stat.c,v $
 * Revision 2.1  90/04/19  20:05:44  jkh
 * Alpha checkin.
 * 
 * Revision 2.0  90/03/26  01:43:42  jkh
 * pre-beta check-in
 *
 */

#include "AwlP.h"
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/stat.h>

/* Return a verbose description of mode */
Local String expand_mode(mode)
ushort mode;
{
     Local char modestr[12];
     int i = 0;

     /*
      * On BSD systems, links are denoted with 'y', on SYSV they're 'l'.
      * Also on BSD a setgid file without execute permission has an 'S'
      * displayed. On SYSV, this means that mandatory locking is set and
      * is indicated with an 'l'. If your system does something else,
      * let me know.
      */
#ifdef BSD
#define LINKCH	'l'
#define NXGID	'S'
#else
#define LINKCH	'y'
#define NXGID	'l'
#endif
     /*
      * figure out the file type, taking various operating system
      * conditionals into account.
      */
     if ((mode & S_IFMT) == S_IFDIR)
	  modestr[0] = 'd';
     else if ((mode & S_IFMT) == S_IFBLK)
	  modestr[0] = 'b';
     else if ((mode & S_IFMT) == S_IFCHR)
	  modestr[0] = 'c';
#ifdef S_IFLNK
     else if ((mode & S_IFMT) == S_IFLNK)
	  modestr[0] = LINKCH;
#endif
#ifdef S_IFIFO
     else if ((mode & S_IFMT) == S_IFIFO)
	  modestr[0] = 'p';
#endif
#ifdef S_IFSOCK
     else if ((mode & S_IFMT) == S_IFSOCK)
	  modestr[0] = 's';
#endif
     else
	  modestr[0] = '-';
     modestr[1] = (mode & (S_IREAD >> 0)) ? 'r' : '-';
     modestr[2] = (mode & (S_IWRITE >> 0)) ? 'w' : '-';
     modestr[3] = (mode & S_ISUID) ? (mode & (S_IEXEC >> 0)) ? 's' : 'S' :
	  (mode & (S_IEXEC >> 0)) ? 'x' : '-';
     modestr[4] = (mode & (S_IREAD >> 3)) ? 'r' : '-';
     modestr[5] = (mode & (S_IWRITE >> 3)) ? 'w' : '-';
     modestr[6] = (mode & S_ISGID) ? (mode & (S_IEXEC >> 3)) ? 's' : NXGID :
	  (mode & (S_IEXEC >> 3)) ? 'x' : '-';
     modestr[7] = (mode & (S_IREAD >> 6)) ? 'r' : '-';
     modestr[8] = (mode & (S_IWRITE >> 6)) ? 'w' : '-';
     modestr[9] = (mode & S_ISVTX) ? (mode & (S_IEXEC >> 3)) ? 't' : 'T' :
	  (mode & (S_IEXEC >> 3)) ? 'x' : '-';
     modestr[10] = '\0';
     return(modestr);
}

Local String file_part(s)
register String s;
{
     register String tmp;

     if (s && *s) {
	  tmp = s + strlen(s);
	  while (tmp != s) {
	       if (*tmp == '/')
		    return(tmp + 1);
	       else
		    --tmp;
	  }
     }
     else
	  tmp = s;
     return(tmp);
}

/*
 * Return an expanded representation of a stat buffer according to a
 * format string.
 */
Export String stat_expand(aw, name, fmt, b)
AwlWidget aw;
String name, fmt;
struct stat *b;
{
     register char ch, op, datefmt[100];
     String ret, tmp;
     int i, max, aux;
     struct group *gw;
     struct passwd *pw;

     ret = NULL;
     i = max = 0;

     if (!fmt)
	  return NULL;

     while (*fmt) {
	  while ((ch = *(fmt++)) != '%') {
	       append_string(&ret, &max, i++, ch);
	       if (ch == '\0')
		    return(ret);
	  }
	  op = (*fmt++);
	  aux = 0;
	  if (*fmt == '(') {
	       ++fmt;
	       while (*fmt && *fmt != ')')
		    datefmt[aux++] = *(fmt++);
	       datefmt[aux] = '\0';
	       if (!*fmt) {
		    exec_warn(aw, "Unterminated date format specifier.");
		    return(NULL);
	       }
	       else
		    ++fmt;	/* skip ')' */
	  }
	  switch (op) {
	  case 'G':	/* group name */
	       if ((gw = getgrgid(b->st_gid)) != NULL)
		    i += sappend_string(&ret, &max, i, gw->gr_name);
	       else
		    i += sappend_string(&ret, &max, i,
					numtoa(aw, b->st_gid, 0));
	       break;

	  case 'P':	/* verbose representation of protection bits */
	       i += sappend_string(&ret, &max, i, expand_mode(b->st_mode));
	       break;

	  case 'U':	/* user name */
	       if ((pw = getpwuid(b->st_uid)) != NULL)
		    i += sappend_string(&ret, &max, i, pw->pw_name);
	       else
		    i += sappend_string(&ret, &max, i,
					numtoa(aw, b->st_uid, 0));
	       break;

	  case 'a':	/* access time */
	       if (aux) {
		    if ((tmp = format_time(datefmt, b->st_atime)) != NULL) {
			 i += sappend_string(&ret, &max, i, tmp);
			 XtFree(tmp);
		    }
		    else
			 i += sappend_string(&ret, &max, i, "<bad date fmt>");
	       }
	       else
		    i += sappend_string(&ret, &max, i,
					numtoa(aw, b->st_atime, 0));
	       break;

#if defined(BSD)
	  case 'b':	/* file size in blocks */
	       i += sappend_string(&ret, &max, i, numtoa(aw, b->st_blocks, 0));
	       break;
#endif

	  case 'c':	/* status change time */
	       if (aux) {
		    if ((tmp = format_time(datefmt, b->st_mtime)) != NULL) {
			 i += sappend_string(&ret, &max, i, tmp);
			 XtFree(tmp);
		    }
		    else
			 i += sappend_string(&ret, &max, i, "<bad date fmt>");
	       }
	       else
		    i += sappend_string(&ret, &max, i,
					numtoa(aw, b->st_ctime, 0));
	       break;

	  case 'd':	/* device */
	       i += sappend_string(&ret, &max, i, numtoa(aw, b->st_dev, 0));
	       break;

	  case 'f':	/* file type */
	       if ((b->st_mode & S_IFMT) == S_IFDIR)
		    append_string(&ret, &max, i++, '/');
#ifdef S_IFLNK
	       else if ((b->st_mode & S_IFMT) == S_IFLNK)
		    append_string(&ret, &max, i++, '@');
#endif
#ifdef S_IFSOCK
	       else if ((b->st_mode & S_IFMT) == S_IFSOCK)
		    append_string(&ret, &max, i++, '=');
#endif
	       else if (b->st_mode & (S_IEXEC|(S_IEXEC >> 3)|(S_IEXEC >> 6)))
		    append_string(&ret, &max, i++, '*');
	       break;

	  case 'g':	/* group ID */
	       i += sappend_string(&ret, &max, i, numtoa(aw, b->st_gid, 0));
	       break;

	  case 'i':	/* inode */
	       i += sappend_string(&ret, &max, i, numtoa(aw, b->st_ino, 0));
	       break;

	  case 'l':	/* link count */
	       i += sappend_string(&ret, &max, i, numtoa(aw, b->st_nlink, 0));
	       break;

	  case 'm':	/* modify time */
	       if (aux) {
		    if ((tmp = format_time(datefmt, b->st_mtime)) != NULL) {
			 i += sappend_string(&ret, &max, i, tmp);
			 XtFree(tmp);
		    }
		    else
			 i += sappend_string(&ret, &max, i, "<bad date fmt>");
	       }
	       else
		    i += sappend_string(&ret, &max, i,
					numtoa(aw, b->st_mtime, 0));
	       break;

	  case 'n':	/* file name */
	       i += sappend_string(&ret, &max, i, file_part(name));
	       break;

	  case 'p':	/* protection mode */
	       i += sappend_string(&ret, &max, i,
				   numtoa(aw, b->st_mode, 0));
	       break;

	  case 'r':
	       i += sappend_string(&ret, &max, i, numtoa(aw, b->st_rdev, 0));
	       break;

	  case 's':	/* file size in bytes */
	       i += sappend_string(&ret, &max, i, numtoa(aw, b->st_size, 0));
	       break;

	  case 't':	/* device type */
	       break;

	  case 'u':	/* user ID */
	       i += sappend_string(&ret, &max, i, numtoa(aw, b->st_uid, 0));
	       break;
#if defined(BSD)
	  case 'z':	/* block size of fs */
	       i += sappend_string(&ret, &max, i,
				   numtoa(aw, b->st_blksize, 0));
	       break;
#endif
	  }
     }
     append_string(&ret, &max, i, '\0');
     return(ret);
}
