#ifndef lint
static char *rcsid = "$Header: util.c,v 2.3 90/04/30 01:13:28 vixie Exp $";
#endif

/*
 *
 *                   Copyright 1989, 1990
 *                    Jordan K. Hubbard
 *
 *		  PCS Computer Systeme, GmbH.
 *	             Munich, West Germany
 *
 *
 * This file is part of AWL.
 *
 * AWL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * AWL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AWL; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

/*
 * Various string/numeric/allocation utilities that don't fit anywhere
 * else.
 *
 * $Log:	util.c,v $
 * Revision 2.3  90/04/30  01:13:28  vixie
 * vixie
 * 
 * Revision 2.2  90/04/19  20:05:59  jkh
 * rcvd
 * 
 * Revision 2.2  90/04/19  20:05:59  jkh
 * Alpha checkin.
 * 
 * Revision 2.1  90/03/29  18:29:05  jkh
 * Various bits of munging around.
 * 
 * Revision 2.0  90/03/26  01:43:57  jkh
 * pre-beta check-in
 * 
 */

#include <sys/time.h>
#include "AwlP.h"
#include "y.tab.h"

Export Inline String skip_whitespace(s)
register String s;
{
     while (*s && isspace(*s))
	  s++;
     return(s);
}

/* Returns the ascii print value for a Datum union, given its type. */
Export String datum_string_value(aw, type, d)
AwlWidget aw;
int type;
Datum d;
{
     Local char chr[2], aval[20];

     switch(type) {
     case CHAR:
	  chr[0] = datum_char(d);
	  chr[1] = '\0';
	  return(chr);

     case INT:
	  return(numtoa(aw, datum_int(d), 0));

     case FLOAT:
	  return(ftoa(datum_float(d)));

     case STRING:
	  return(datum_string(d));

     case RANGE:
	  sprintf(aval, "%d..%d", range_min(datum_range(d)),
		  range_max(datum_range(d)));
	  return(aval);

     default:
	  sprintf(aval, "%#08x", datum_any(d));
	  return(aval);
     }
}

/* Same as above but for values (shorthand) */
Export String value_string_value(aw, val)
AwlWidget aw;
Value val;
{
     return(datum_string_value(aw, value_type(val), value_data(val)));
}

/* format a time specifier and return it */
Export String format_time(fmt, t)
String fmt;
time_t t;
{
     char buffer[512];

     if (strftime(buffer, 512, fmt, localtime(&t)))
	  return(XtNewString(buffer));
     else
	  return(NULL);
}

/*
 * "Safe" strcmp. Handles NULL pointer value for either string.
 */
Export int strcomp(s1, s2)
register String s1, s2;
{
     if (s1 && s2)
	  return(strcmp(s1, s2));
     else if (!s1 && !s2)
	  return(0);
     else if (!s1 && s2)
	  return(-1);
     else
	  return(1);
}

/*
 * Another "safe" strcmp that ignores case.
 */
Export int stricomp(s1, s2)
register String s1, s2;
{
#ifndef tolower
     Import char tolower();
#endif

     if (s1 && s2) {
          if (strlen(s1) != strlen(s2))
              return(-1);

          while (*s1 && *s2 && (tolower(*s1) == tolower(*s2)))
               s1++, s2++;
          if (!*s1 && !*s2)
               return(0);
          else if (*s1 < *s2)
               return (-1);
          else
               return(1);
     }
     else if (!s1 && !s2)
          return(0);
     else if (!s1 && s2)
          return(-1);
     else
          return(1);
}

/* list manipulation routines */

/* free all elements of a list */
Export void list_free(l)
register String *l;
{
     if (l)
	  while (*l)
	       XtFree(*(l++));
}

/* return the number of elements in a list */
Export int list_len(l)
register String *l;
{
     register int i = 0;

     if (l)
	  while (*(l++))
	       ++i;
     return(i);
}

/* indicate whether or not all elements of a list are the same length */
Export int list_is_rect(l)
register String *l;
{
     register int i, comp, len = list_len(l);

     if (len) {
	  comp = strlen(l[0]);
	  for (i = 1; i < len; i++)
	       if (strlen(l[i]) != comp)
		    return 0;
	  return 1;
     }
     return 0;
}
  
Export String *list_dup(aw, l)
AwlWidget aw;
register String *l;
{
     register int i, j;
     register String *m = NULL;

     if (i = list_len(l)) {
	  m = (String *)XtMalloc((i + 1) * sizeof(String *));
	  for (j = 0; j < i; j++) {
	       String tmp;
	       int len = strlen(l[j]);

	       m[j] = XtMalloc(len + 1);
	       strncpy(m[j], l[j], len);
	       m[j][len] = '\0';
	  }
	  m[j] = NULL;
     }
     return(m);
}

#define BINGO '\377'

Export String *list_from_string(aw, s)
AwlWidget aw;
register String s;
{
     String *ret;

     if (!s)
	 ret = NULL;
     else {
	 register int zap_char;
	 register int len = strlen(s);
	 register int count;
	 register int i;
	 register String tmp, strsed_cmd;
	 int sep_len = strlen(awl_sep(aw));
	 
	 switch (sep_len){
	     case 0:
		 /*
		  * Make a list with an element for each char in s.
		  */
		 tmp = (String)XtMalloc(len * 2);
		 ret = (String *)XtMalloc((len + 1) * sizeof(String));
		 for (i = 0; i < len; i++){
		     tmp[2 * i] = s[i];
		     tmp[2 * i + 1] = '\0';
		     ret[i] = tmp + 2 * i;
		 } 
		 ret[len] = NULL;
		 return ret;
		 break;
	     
	     case 1:
		 /*
		  * Do it the fast way - no strsed here.
		  */
		 tmp = (String)XtMalloc(len + 1);
		 zap_char = awl_sep(aw)[0];
		 tmp[0] = '\0';
		 strcat(tmp, s);
		 break;
	     
	     default:
		 /*
		  * Use strsed to zap the regex into an "unlikely"
		  * char. Then proceed as though that char (BINGO) 
		  * were the actual separator.
		  */
		 strsed_cmd = (String)XtMalloc(6 + sep_len);
		 sprintf(strsed_cmd, "g/%s/%c/", awl_sep(aw), BINGO);
		 if (!(tmp = strsed(s, strsed_cmd))){
		     XtFree(strsed_cmd);
		     return NULL;
		 }
		 XtFree(strsed_cmd);
		 len = strlen(tmp);
		 zap_char = BINGO;
		 break;
	     
	 }
	 
	 if (tmp[len - 1] == zap_char)
	     len--;
	 
	 count = 1;
	 for (i = 0; i < len; i++)
	     if (tmp[i] == zap_char){
		 tmp[i] = '\0';
		 count++;
	     }
	 
	 tmp[len] = '\0';
	 
	 ret = (String *)XtMalloc((count + 1) * sizeof(String));
	 
	 for (i = 0; i < count; i++){
	     ret[i] = tmp;
	     while (*tmp++)
		 ;
	 }
	 ret[count] = NULL;
     }
     return(ret);
}

#undef BINGO


Export String list_to_string(aw, l)
AwlWidget aw;
register String *l;
{
     register String ret = NULL;

     if (l) {
	 register int items, i, len, sep_len;
	 register String sep = awl_sep(aw);
	 
	 items = i = len = 0;
	 while (l[items]) {
	     len += strlen(l[items]);
	     items++;
	 }
	 sep_len = strlen(sep);
	 ret = (String)XtMalloc(len + (items * sep_len) + 1);
	 i = len = 0;
	 
	 /*
	  * Copy the first items - 1 of them, each followed by a separator.
	  */
	 for (i = 0; i < items; i++) {
	     strcpy(ret + len, l[i]);
	     len += strlen(l[i]);
	     strcpy(ret + len, sep);
	     len += sep_len;
	 }
	 ret[len] = '\0';
     }
     return(ret);
}

#define LIST_INCR	10

struct _linfo {
     String *l;
     int max, idx;
};

Local int build_list(sym, ll)
Symbol sym;
register struct _linfo *ll;
{
     if (ll->idx == (ll->max - 1)) {
	  ll->max += LIST_INCR;
	  ll->l = (String *)XtRealloc(ll->l, ll->max * sizeof(String *));
     }
     else
	  ll->l[ll->idx++] = XtNewString(symbol_name(sym));
     return(TRUE);
}

Export String *list_from_alist(aw, t)
AwlWidget aw;
Table t;
{
     struct _linfo li;

     li.idx = 0;
     li.max = LIST_INCR;
     li.l = (String *)XtMalloc(li.max * sizeof(String *));
     hash_traverse(t, build_list, &li);
     li.l[li.idx] = NULL;
     return(li.l);
}

/*
 * Do a "search only" strsed, returning only the portion of string
 * that matched.
 */
Export String s_strsed(s1, s2)
String s1, s2;
{
     register String tmp, tmp1;
     int r[2];
     Boolean noslash = FALSE;

     if (!s1 || !s2)
	  return(NULL);
     else if (s2[0] != '/') {
	  tmp1 = XtMalloc(strlen(s2) + 3);
	  tmp1[0] = '/'; tmp1[1] = '\0';
	  strcat(tmp1, s2);
	  strcat(tmp1, "/");
	  noslash = TRUE;
     }
     else
	  tmp1 = s2;
     if ((tmp = strsed(s1, tmp1, r)) != NULL) {
	  register int i = 0;
	  register String s = NULL;

	  if (r[1] > 0) {
	       s = XtMalloc((r[1] - r[0]) + 1);
	       
	       while (r[0] < r[1])
		    s[i++] = tmp[r[0]++];
	       s[i] = '\0';
	  }
	  XtFree(tmp);
	  tmp = s;
     }
     if (noslash)
	  XtFree(tmp1);
     return(tmp);
}

/* Append a character to a string, lengthening it if necessary. */
Export void append_string(sp, lp, idx, ch)
String *sp;
int *lp, idx;
char ch;
{
     if (!*lp) {
	  *lp = ALLOC_INCR;
	  *sp = XtMalloc(ALLOC_INCR);
     }
     else if (idx >= *lp) {
	  *lp += ALLOC_INCR;
	  *sp = XtRealloc(*sp, *lp);
     }
     (*sp)[idx] = ch;
}

/* Append a string to another string, lengthening target if necessary. */
Export int sappend_string(sp, lp, idx, s)
String *sp;
int *lp, idx;
String s;
{
     int len;

     len = strlen(s);
     if (!*lp) {
	  *lp = ALLOC_INCR;
	  *sp = XtMalloc(ALLOC_INCR);
     }
     else if (idx + len >= *lp) {
	  if (len < ALLOC_INCR)
	       *lp += ALLOC_INCR;
	  else /* round up */
	       *lp += len + (len % ALLOC_INCR);
	  *sp = XtRealloc(*sp, *lp);
     }
     strcpy(*sp + idx, s);
     return(len);
}

/* Create a string that's a multiple of itself */
Export String string_multiply(str, m)
register String str;
register int m;
{
     register String ret;
     register int i, o, l = strlen(str);

     ret = XtMalloc((l * m) + 1);
     o = 0;
     for (i = 0; i < m; i++) {
	  strcpy(ret + o, str);
	  o += l;
     }
     ret[o] = '\0';
     return(ret);
}

Local void zombie_time()
{
     signal(SIGALRM, SIG_DFL);
}

/* process file descriptor routines */
Export Process p_popen(aw, s, mode)
AwlWidget aw;
String s, mode;
{
     Process ret = NULL;
     int in[2], out[2], pid;

     if (pipe(in) < 0 || pipe(out) < 0)
	  exec_warn(aw, "p_popen: Couldn't open input or output pipe");
     else {
	  ret = (Process)XtMalloc(process_size);
	  if (proc_pid(ret) = fork()) {	/* der Vati */
	       if (*mode == 'r' || *mode == '+')
		    proc_in(ret) = fdopen(in[0], "r");
	       else {
		    proc_in(ret) = NULL;
		    close(in[0]);
	       }
	       if (*mode == 'w' || *mode == '+')
		    proc_out(ret) = fdopen(out[1], "w");
	       else {
		    proc_out(ret) = NULL;
		    close(out[1]);
	       }
	       close(in[1]);
	       close(out[0]);
	  }
	  else {			/* das Kind */
	       int fl;

	       if (*mode == 'w' || *mode == '+')
		    dup2(out[0], 0);
	       else
		    close(out[0]);
	       if (*mode == 'r' || *mode == '+') {
		    dup2(in[1], 1);
		    dup2(in[1], 2);
	       }
	       else
		    close(in[1]);

	       for (fl = 3; fl < NFILES; fl++)
		    close(fl);

	       setpgrp(0, getpid());
	       execl("/bin/sh", "sh", "-c", s, 0);
	       exec_warn(aw, "Couldn't exec shell on '%s'", s);
	  }
     }
     return(ret);
}

Export int p_pclose(aw, p)
AwlWidget aw;
Process p;
{
     int ret, pid, status;

     pid = proc_pid(p);
     if (proc_in(p))
	  fclose(proc_in(p));
     if (proc_out(p))
	  fclose(proc_out(p));
     XtFree(p);

     /*
      * Set an ALARM to prevent really bad process states from wedging the
      * whole show.
      */
     signal(SIGALRM, zombie_time);
     alarm(20);
     while ((ret = wait(&status)) != pid && ret > 0);
     if (ret < 0) { /* alarm went off */
          exec_warn(aw, "p_pclose: Abandoned wait() for wedged process.");
	  /* Try nuking it as a parting shot */
	  kill(pid, SIGKILL);
	  status = -1;
     }
     else {
	  alarm(0);
	  signal(SIGALRM, SIG_DFL);
     }
     return(status);
}

/* Allocation routines */

Export Inline AllocObj new_aobj(ptr)
Generic ptr;
{
     AllocObj ret;

     ret = (AllocObj)XtMalloc(aobj_size);
     aobj_cnt(ret) = 0;
     aobj_ptr(ret) = ptr;
     return(ret);
}

Export Layout new_wl(type)
WidgetClass type;
{
     Layout tmp;

     tmp = (Layout)XtMalloc(wl_size);
     bzero(tmp, wl_size);
     wl_wclass(tmp) = type;
     return(tmp);
}

Export Resource new_res(p)
Layout p;
{
     Resource tmp;

     tmp = (Resource)XtMalloc(res_size);
     bzero(tmp, res_size);
     res_parent(tmp) = p;
     return(tmp);
}

Export Breakpoint new_bkpt(addr)
int addr;
{
     Breakpoint tmp;

     tmp = (Breakpoint)XtMalloc(bkpt_size);
     bzero(tmp, bkpt_size);
     bkpt_addr(tmp) = addr;
     return(tmp);
}
